"""OpenTelemetry Groq instrumentation"""

import logging
from typing import Collection

from opentelemetry import context as context_api
from .config import Config
from .span_utils import (
    set_input_attributes,
    set_model_input_attributes,
    set_model_response_attributes,
    set_model_streaming_response_attributes,
    set_response_attributes,
    set_streaming_response_attributes,
)
from .version import __version__

from lmnr.opentelemetry_lib.opentelemetry.instrumentation.shared.utils import (
    dont_throw,
    safe_start_span,
)
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry.instrumentation.utils import _SUPPRESS_INSTRUMENTATION_KEY, unwrap
from opentelemetry.trace import Tracer, get_tracer
from opentelemetry.trace.status import Status, StatusCode
from wrapt import wrap_function_wrapper

from groq._streaming import AsyncStream, Stream

logger = logging.getLogger(__name__)

_instruments = ("groq >= 0.9.0",)


WRAPPED_METHODS = [
    {
        "package": "groq.resources.chat.completions",
        "object": "Completions",
        "method": "create",
        "span_name": "groq.chat",
    },
]
WRAPPED_AMETHODS = [
    {
        "package": "groq.resources.chat.completions",
        "object": "AsyncCompletions",
        "method": "create",
        "span_name": "groq.chat",
    },
]


def is_streaming_response(response):
    return isinstance(response, Stream) or isinstance(response, AsyncStream)


def _with_chat_telemetry_wrapper(func):
    """Helper for providing tracer for wrapper functions. Includes metric collectors."""

    def _with_chat_telemetry(
        tracer,
        to_wrap,
    ):
        def wrapper(wrapped, instance, args, kwargs):
            return func(
                tracer,
                to_wrap,
                wrapped,
                instance,
                args,
                kwargs,
            )

        return wrapper

    return _with_chat_telemetry


def _process_streaming_chunk(chunk):
    """Extract content, finish_reason and usage from a streaming chunk."""
    if not chunk.choices:
        return None, None, None

    delta = chunk.choices[0].delta
    content = delta.content if hasattr(delta, "content") else None
    finish_reason = chunk.choices[0].finish_reason

    # Extract usage from x_groq if present in the final chunk
    usage = None
    if hasattr(chunk, "x_groq") and chunk.x_groq and chunk.x_groq.usage:
        usage = chunk.x_groq.usage

    return content, finish_reason, usage


@dont_throw
def _handle_streaming_response(span, accumulated_content, finish_reason, usage):
    set_model_streaming_response_attributes(span, usage)
    set_streaming_response_attributes(span, accumulated_content, finish_reason, usage)


def _create_stream_processor(
    response,
    span,
):
    """Create a generator that processes a stream while collecting telemetry."""
    accumulated_content = ""
    finish_reason = None
    usage = None

    for chunk in response:
        try:
            content, chunk_finish_reason, chunk_usage = _process_streaming_chunk(chunk)
            if content:
                accumulated_content += content
            if chunk_finish_reason:
                finish_reason = chunk_finish_reason
            if chunk_usage:
                usage = chunk_usage
        except Exception as e:
            logger.warning(
                "Failed to process streaming chunk for groq span, error: %s", str(e)
            )
        finally:
            yield chunk

    _handle_streaming_response(span, accumulated_content, finish_reason, usage)

    if span.is_recording():
        span.set_status(Status(StatusCode.OK))

    span.end()


async def _create_async_stream_processor(response, span):
    """Create an async generator that processes a stream while collecting telemetry."""
    accumulated_content = ""
    finish_reason = None
    usage = None

    async for chunk in response:
        try:
            content, chunk_finish_reason, chunk_usage = _process_streaming_chunk(chunk)
            if content:
                accumulated_content += content
            if chunk_finish_reason:
                finish_reason = chunk_finish_reason
            if chunk_usage:
                usage = chunk_usage
        except Exception as e:
            logger.warning(
                "Failed to process streaming chunk for groq span, error: %s", str(e)
            )
        finally:
            yield chunk

    _handle_streaming_response(span, accumulated_content, finish_reason, usage)

    if span.is_recording():
        span.set_status(Status(StatusCode.OK))

    span.end()


@dont_throw
def _handle_input(span, kwargs):
    set_model_input_attributes(span, kwargs)
    set_input_attributes(span, kwargs)


@dont_throw
def _handle_response(span, response):
    set_model_response_attributes(span, response)
    set_response_attributes(span, response)


@_with_chat_telemetry_wrapper
def _wrap(
    tracer: Tracer,
    to_wrap,
    wrapped,
    instance,
    args,
    kwargs,
):
    """Instruments and calls every function defined in TO_WRAP."""
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
        return wrapped(*args, **kwargs)

    name = to_wrap.get("span_name")
    span = safe_start_span(
        name=name, attributes={"gen_ai.system": "groq"}, span_type="LLM"
    )
    if not span:
        logger.warning("Failed to start span for groq chat")
        return wrapped(*args, **kwargs)

    _handle_input(span, kwargs)

    try:
        response = wrapped(*args, **kwargs)
    except Exception as e:  # pylint: disable=broad-except
        raise e

    if is_streaming_response(response):
        try:
            return _create_stream_processor(response, span)
        except Exception as ex:
            logger.warning(
                "Failed to process streaming response for groq span, error: %s",
                str(ex),
            )
            span.set_status(Status(StatusCode.ERROR))
            span.end()
            raise
    elif response:
        try:
            _handle_response(span, response)

        except Exception as ex:  # pylint: disable=broad-except
            logger.warning(
                "Failed to set response attributes for groq span, error: %s",
                str(ex),
            )

        if span.is_recording():
            span.set_status(Status(StatusCode.OK))
    span.end()
    return response


@_with_chat_telemetry_wrapper
async def _awrap(
    tracer,
    to_wrap,
    wrapped,
    instance,
    args,
    kwargs,
):
    """Instruments and calls every function defined in TO_WRAP."""
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
        return await wrapped(*args, **kwargs)

    name = to_wrap.get("span_name")
    span = safe_start_span(
        name=name, attributes={"gen_ai.system": "groq"}, span_type="LLM"
    )
    if not span:
        logger.warning("Failed to start span for groq chat")
        return await wrapped(*args, **kwargs)

    _handle_input(span, kwargs)

    try:
        response = await wrapped(*args, **kwargs)
    except Exception as e:  # pylint: disable=broad-except
        raise e

    if is_streaming_response(response):
        try:
            return await _create_async_stream_processor(response, span)
        except Exception as ex:
            logger.warning(
                "Failed to process streaming response for groq span, error: %s",
                str(ex),
            )
            span.set_status(Status(StatusCode.ERROR))
            span.end()
            raise
    elif response:
        _handle_response(span, response)

        if span.is_recording():
            span.set_status(Status(StatusCode.OK))
    span.end()
    return response


class GroqInstrumentor(BaseInstrumentor):
    """An instrumentor for Groq's client library."""

    def __init__(
        self,
        enrich_token_usage: bool = False,
        use_legacy_attributes: bool = True,
    ):
        super().__init__()
        Config.enrich_token_usage = enrich_token_usage
        Config.use_legacy_attributes = use_legacy_attributes

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        tracer_provider = kwargs.get("tracer_provider")
        tracer = get_tracer(__name__, __version__, tracer_provider)

        for wrapped_method in WRAPPED_METHODS:
            wrap_package = wrapped_method.get("package")
            wrap_object = wrapped_method.get("object")
            wrap_method = wrapped_method.get("method")

            try:
                wrap_function_wrapper(
                    wrap_package,
                    f"{wrap_object}.{wrap_method}",
                    _wrap(
                        tracer,
                        wrapped_method,
                    ),
                )
            except ModuleNotFoundError:
                pass  # that's ok, we don't want to fail if some methods do not exist

        for wrapped_method in WRAPPED_AMETHODS:
            wrap_package = wrapped_method.get("package")
            wrap_object = wrapped_method.get("object")
            wrap_method = wrapped_method.get("method")
            try:
                wrap_function_wrapper(
                    wrap_package,
                    f"{wrap_object}.{wrap_method}",
                    _awrap(
                        tracer,
                        wrapped_method,
                    ),
                )
            except ModuleNotFoundError:
                pass  # that's ok, we don't want to fail if some methods do not exist

    def _uninstrument(self, **kwargs):
        for wrapped_method in WRAPPED_METHODS:
            wrap_package = wrapped_method.get("package")
            wrap_object = wrapped_method.get("object")
            unwrap(
                f"{wrap_package}.{wrap_object}",
                wrapped_method.get("method"),
            )
        for wrapped_method in WRAPPED_AMETHODS:
            wrap_package = wrapped_method.get("package")
            wrap_object = wrapped_method.get("object")
            unwrap(
                f"{wrap_package}.{wrap_object}",
                wrapped_method.get("method"),
            )



import logging

from opentelemetry.semconv._incubating.attributes.gen_ai_attributes import (
    GEN_AI_RESPONSE_ID,
    GEN_AI_RESPONSE_MODEL,
    GEN_AI_USAGE_INPUT_TOKENS,
    GEN_AI_USAGE_OUTPUT_TOKENS,
)
from opentelemetry.trace.status import Status, StatusCode

from lmnr.sdk.utils import json_dumps

from .config import Config
from .span_utils import (
    set_streaming_response_attributes,
)
from .utils import (
    dont_throw,
    set_span_attribute,
)

logger = logging.getLogger(__name__)


@dont_throw
def _process_response_item(item, complete_response):
    if item.type == "message_start":
        complete_response["model"] = item.message.model
        usage = dict(item.message.usage)
        complete_response["usage"] = usage
        complete_response["service_tier"] = usage.get("service_tier") or None
        complete_response["id"] = item.message.id
    elif item.type == "content_block_start":
        index = item.index
        if len(complete_response.get("events")) <= index:
            complete_response["events"].append(
                {"index": index, "text": "", "type": item.content_block.type}
            )
            if item.content_block.type == "tool_use":
                complete_response["events"][index]["id"] = item.content_block.id
                complete_response["events"][index]["name"] = item.content_block.name
                complete_response["events"][index]["input"] = ""

    elif item.type == "content_block_delta":
        index = item.index
        if item.delta.type == "thinking_delta":
            complete_response["events"][index]["text"] += item.delta.thinking or ""
        elif item.delta.type == "text_delta":
            complete_response["events"][index]["text"] += item.delta.text or ""
        elif item.delta.type == "input_json_delta":
            complete_response["events"][index]["input"] += item.delta.partial_json
    elif item.type == "message_delta":
        for event in complete_response.get("events", []):
            event["finish_reason"] = item.delta.stop_reason
        if item.usage:
            # message_delta usage values are cumulative (per Anthropic docs),
            # so we update/replace rather than add to existing values.
            # Filter out None values to avoid overwriting message_start data
            # (e.g. cache_creation_input_tokens) with None from message_delta.
            usage_update = {
                k: v for k, v in dict(item.usage).items() if v is not None
            }
            if "usage" in complete_response:
                complete_response["usage"].update(usage_update)
            else:
                complete_response["usage"] = usage_update
    elif item.type in ["message_stop", "message_start"]:
        # raw stream returns the service_tier in the message_start event
        # messages.stream returns the service_tier in the message_stop event
        usage = dict(item.message.usage or {})
        complete_response["service_tier"] = usage.get("service_tier")


def _set_token_usage(
    span,
    complete_response,
    prompt_tokens,
    completion_tokens,
):
    cache_read_tokens = (
        complete_response.get("usage", {}).get("cache_read_input_tokens", 0) or 0
    )
    cache_creation_tokens = (
        complete_response.get("usage", {}).get("cache_creation_input_tokens", 0) or 0
    )

    input_tokens = prompt_tokens + cache_read_tokens + cache_creation_tokens

    set_span_attribute(span, GEN_AI_USAGE_INPUT_TOKENS, input_tokens)
    set_span_attribute(span, GEN_AI_USAGE_OUTPUT_TOKENS, completion_tokens)

    set_span_attribute(span, GEN_AI_RESPONSE_MODEL, complete_response.get("model"))
    set_span_attribute(span, "gen_ai.usage.cache_read_input_tokens", cache_read_tokens)
    set_span_attribute(
        span,
        "gen_ai.usage.cache_creation_input_tokens",
        cache_creation_tokens,
    )


def _handle_streaming_response(span, complete_response, record_raw_response=False):
    if not span.is_recording():
        return
    result = set_streaming_response_attributes(span, complete_response.get("events"))

    if record_raw_response and result:
        try:
            # Enrich the result with static attributes
            result["id"] = complete_response.get("id")
            result["model"] = complete_response.get("model")
            result["type"] = "message"
            result["usage"] = complete_response.get("usage") or {
                "input_tokens": 0,
                "output_tokens": 0,
            }

            set_span_attribute(span, "lmnr.sdk.raw.response", json_dumps(result))
        except Exception:
            pass


@dont_throw
def build_from_streaming_response(
    span,
    response,
    instance,
    kwargs: dict = {},
    record_raw_response: bool = False,
):
    complete_response = {
        "events": [],
        "model": "",
        "usage": {},
        "id": "",
        "service_tier": None,
    }

    for item in response:
        try:
            yield item
        except Exception as e:
            raise e
        _process_response_item(item, complete_response)

    set_span_attribute(span, GEN_AI_RESPONSE_ID, complete_response.get("id"))
    set_span_attribute(
        span,
        "anthropic.response.service_tier",
        complete_response.get("service_tier"),
    )

    try:
        usage = complete_response.get("usage")
        prompt_tokens = (usage.get("input_tokens", 0) or 0) if usage else 0
        completion_tokens = (usage.get("output_tokens", 0) or 0) if usage else 0

        if not usage and Config.enrich_token_usage:
            completion_content = ""
            if complete_response.get("events"):
                model_name = complete_response.get("model") or None
                for event in complete_response.get("events") or []:
                    if event.get("text"):
                        completion_content += event.get("text")

                if model_name and hasattr(instance, "count_tokens"):
                    completion_tokens = instance.count_tokens(completion_content)

        _set_token_usage(
            span,
            complete_response,
            prompt_tokens,
            completion_tokens,
        )
    except Exception as e:
        logger.warning("Failed to set token usage, error: %s", e)

    _handle_streaming_response(span, complete_response, record_raw_response)

    if span.is_recording():
        span.set_status(Status(StatusCode.OK))
        span.end()


@dont_throw
async def abuild_from_streaming_response(
    span,
    response,
    instance,
    kwargs: dict = {},
    record_raw_response: bool = False,
):
    complete_response = {
        "events": [],
        "model": "",
        "usage": {},
        "id": "",
        "service_tier": None,
    }
    async for item in response:
        try:
            yield item
        except Exception as e:
            raise e
        _process_response_item(item, complete_response)

    set_span_attribute(span, GEN_AI_RESPONSE_ID, complete_response.get("id"))
    set_span_attribute(
        span,
        "anthropic.response.service_tier",
        complete_response.get("service_tier"),
    )

    try:
        usage = complete_response.get("usage")
        prompt_tokens = (usage.get("input_tokens", 0) or 0) if usage else 0
        completion_tokens = (usage.get("output_tokens", 0) or 0) if usage else 0

        if not usage and Config.enrich_token_usage:
            completion_content = ""
            if complete_response.get("events"):
                model_name = complete_response.get("model") or None
                for event in complete_response.get("events") or []:
                    if event.get("text"):
                        completion_content += event.get("text")

                if model_name and hasattr(instance, "count_tokens"):
                    completion_tokens = instance.count_tokens(completion_content)

        _set_token_usage(
            span,
            complete_response,
            prompt_tokens,
            completion_tokens,
        )
    except Exception as e:
        logger.warning("Failed to set token usage, error: %s", str(e))

    _handle_streaming_response(span, complete_response, record_raw_response)

    if span.is_recording():
        span.set_status(Status(StatusCode.OK))
        span.end()


class WrappedMessageStreamManager:
    """Wrapper for MessageStreamManager that handles instrumentation"""

    def __init__(
        self,
        stream_manager,
        span,
        instance,
        kwargs,
        record_raw_response: bool = False,
    ):
        self._stream_manager = stream_manager
        self._span = span
        self._instance = instance
        self._kwargs = kwargs
        self._record_raw_response = record_raw_response

    def __enter__(self):
        # Call the original stream manager's __enter__ to get the actual stream
        stream = self._stream_manager.__enter__()
        # Return the wrapped stream
        return build_from_streaming_response(
            self._span,
            stream,
            self._instance,
            self._kwargs,
            record_raw_response=self._record_raw_response,
        )

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self._stream_manager.__exit__(exc_type, exc_val, exc_tb)


class WrappedAsyncMessageStreamManager:
    """Wrapper for AsyncMessageStreamManager that handles instrumentation"""

    def __init__(
        self,
        stream_manager,
        span,
        instance,
        kwargs,
        record_raw_response: bool = False,
    ):
        self._stream_manager = stream_manager
        self._span = span
        self._instance = instance
        self._kwargs = kwargs
        self._record_raw_response = record_raw_response

    async def __aenter__(self):
        # Call the original stream manager's __aenter__ to get the actual stream
        stream = await self._stream_manager.__aenter__()
        # Return the wrapped stream
        return abuild_from_streaming_response(
            self._span,
            stream,
            self._instance,
            self._kwargs,
            record_raw_response=self._record_raw_response,
        )

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        return await self._stream_manager.__aexit__(exc_type, exc_val, exc_tb)



"""Shared utilities for Claude Agent instrumentation."""

import os
import re
import socket
import time

from lmnr.sdk.log import get_default_logger

logger = get_default_logger(__name__)
# Constants
DEFAULT_ANTHROPIC_BASE_URL = "https://api.anthropic.com"
FOUNDRY_BASE_URL_ENV = "ANTHROPIC_FOUNDRY_BASE_URL"
FOUNDRY_RESOURCE_ENV = "ANTHROPIC_FOUNDRY_RESOURCE"
FOUNDRY_USE_ENV = "CLAUDE_CODE_USE_FOUNDRY"
BEDROCK_BASE_URL_ENV = "ANTHROPIC_BEDROCK_BASE_URL"
BEDROCK_USE_ENV = "CLAUDE_CODE_USE_BEDROCK"
BEDROCK_AWS_REGION_ENV = "AWS_REGION"


def is_truthy_env(value: str | None) -> bool:
    """Check if environment variable value is truthy (equals '1')."""
    return value == "1"


def snapshot_env(keys: list[str]) -> tuple[dict[str, str | None], set[str]]:
    """
    Snapshot environment variables.

    Returns:
        Tuple of (snapshot dict, set of keys that were present)
    """
    snapshot: dict[str, str | None] = {}
    set_keys: set[str] = set()
    for key in keys:
        if key in os.environ:
            set_keys.add(key)
            snapshot[key] = os.environ.get(key)
        else:
            snapshot[key] = None
    return snapshot, set_keys


def restore_env(snapshot: dict[str, str | None], set_keys: set[str]) -> None:
    """
    Restore environment variables from snapshot.

    Args:
        snapshot: Dictionary of variable names to values
        set_keys: Set of keys that were originally present
    """
    for key, value in snapshot.items():
        if key in set_keys:
            os.environ[key] = value if value is not None else ""
        else:
            os.environ.pop(key, None)


def is_port_open(port: int, timeout: float = 0.5) -> bool:
    """
    Check if a port is currently accepting connections.

    Args:
        port: Port number to check
        timeout: Connection timeout in seconds

    Returns:
        True if port is open, False otherwise
    """
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        sock.settimeout(timeout)
        try:
            sock.connect(("127.0.0.1", port))
            return True
        except OSError:
            return False


def wait_for_port(port: int, timeout: float = 5.0) -> bool:
    """
    Wait for a port to become available (accepting connections).

    Args:
        port: Port number to wait for
        timeout: Maximum time to wait in seconds

    Returns:
        True if port became available, False if timeout
    """
    deadline = time.monotonic() + timeout
    while time.monotonic() < deadline:
        if is_port_open(port, timeout=0.2):
            return True
        time.sleep(0.1)
    return False


def _get_region_from_aws_config(profile: str) -> str | None:
    """Read region for a given profile from ~/.aws/config."""
    config_path = os.path.expanduser("~/.aws/config")
    try:
        with open(config_path, "r") as f:
            content = f.read()
    except OSError:
        return None

    profile_header = "default" if profile == "default" else f"profile {profile}"
    match = re.search(
        rf"\[{re.escape(profile_header)}\][^\[]*?^\s*region\s*=\s*([^\s\n]+)",
        content,
        re.MULTILINE | re.DOTALL,
    )
    return match.group(1) if match else None


def resolve_target_url_from_env(
    env_dict: dict[str, str], fallback: str = DEFAULT_ANTHROPIC_BASE_URL
) -> str | None:
    """
    Resolve target URL from environment dictionary with os.environ fallback.

    This is the single source of truth for determining the target URL for the proxy.

    Resolution order (highest to lowest priority):
    1. HTTPS_PROXY - if set, use as target (our proxy will forward to it)
    2. HTTP_PROXY - if set, use as target (our proxy will forward to it)
    3. Third-party provider URLs (e.g., Foundry, Bedrock):
       - If CLAUDE_CODE_USE_FOUNDRY is truthy:
         - Use ANTHROPIC_FOUNDRY_BASE_URL, or
         - Construct from ANTHROPIC_FOUNDRY_RESOURCE
       - If CLAUDE_CODE_USE_BEDROCK is truthy:
         - Use ANTHROPIC_BEDROCK_BASE_URL, or
         - Construct from AWS_REGION env var, or
         - Construct by reading region from ~/.aws/config via AWS_PROFILE
    4. ANTHROPIC_BASE_URL - standard Anthropic API base URL
    5. Fall back to default (https://api.anthropic.com)

    For each environment variable, checks env_dict first, then os.environ as fallback.

    Args:
        env_dict: Dictionary of environment variables (e.g., from options.env)
        fallback: Fallback URL if no other source found (default: DEFAULT_ANTHROPIC_BASE_URL)

    Returns:
        Resolved target URL, or None if provider is misconfigured
    """

    # Helper to get value from env_dict first, then os.environ
    def get_env_value(key: str) -> str | None:
        return env_dict.get(key) or os.environ.get(key)

    # 1. Check for HTTPS_PROXY (highest priority)
    https_proxy = get_env_value("HTTPS_PROXY")
    if https_proxy:
        return https_proxy.rstrip("/")

    # 2. Check for HTTP_PROXY
    http_proxy = get_env_value("HTTP_PROXY")
    if http_proxy:
        return http_proxy.rstrip("/")

    # 3. Check for third-party providers (Foundry)
    foundry_enabled = is_truthy_env(get_env_value(FOUNDRY_USE_ENV))
    if foundry_enabled:
        # Try to get Foundry base URL first
        foundry_base_url = get_env_value(FOUNDRY_BASE_URL_ENV)
        if foundry_base_url:
            return foundry_base_url.rstrip("/")

        # Try to construct from resource
        foundry_resource = get_env_value(FOUNDRY_RESOURCE_ENV)
        if foundry_resource:
            return f"https://{foundry_resource}.services.ai.azure.com/anthropic"

        # Foundry is enabled but misconfigured
        logger.error(
            "%s is set but neither %s nor %s is configured. "
            "Microsoft Foundry requires one of these values.",
            FOUNDRY_USE_ENV,
            FOUNDRY_BASE_URL_ENV,
            FOUNDRY_RESOURCE_ENV,
        )
        return None

    # 3b. Check for Bedrock
    bedrock_enabled = is_truthy_env(get_env_value(BEDROCK_USE_ENV))
    if bedrock_enabled:
        bedrock_base_url = get_env_value(BEDROCK_BASE_URL_ENV)
        if bedrock_base_url:
            return bedrock_base_url.rstrip("/")

        region = get_env_value(BEDROCK_AWS_REGION_ENV)
        if not region:
            aws_profile = get_env_value("AWS_PROFILE") or "default"
            region = _get_region_from_aws_config(aws_profile)

        if region:
            return f"https://bedrock-runtime.{region}.amazonaws.com"

        logger.error(
            "%s is set but could not determine AWS region. "
            "Set %s or configure a region in ~/.aws/config for the active profile.",
            BEDROCK_USE_ENV,
            BEDROCK_AWS_REGION_ENV,
        )
        return None

    # 4. Check for ANTHROPIC_BASE_URL
    anthropic_base_url = get_env_value("ANTHROPIC_BASE_URL")
    if anthropic_base_url:
        return anthropic_base_url.rstrip("/")

    # 5. Use fallback
    return fallback


def setup_proxy_env(proxy_url: str) -> dict[str, str | None]:
    """
    Configure global environment to use proxy for custom transports.

    This is only used for custom (non-SubprocessCLITransport) transports
    where we can't control environment variable passing. We set ANTHROPIC_ORIGINAL_BASE_URL
    so the proxy server knows where to forward requests.

    Also removes HTTP_PROXY and HTTPS_PROXY from global env
    since our proxy will handle forwarding to them.

    Args:
        proxy_url: Proxy base URL (e.g., "http://127.0.0.1:45667")

    Returns:
        Dictionary of original env values for restoration
    """
    snapshot: dict[str, str | None] = {
        "ANTHROPIC_BASE_URL": os.environ.get("ANTHROPIC_BASE_URL"),
        "ANTHROPIC_ORIGINAL_BASE_URL": os.environ.get("ANTHROPIC_ORIGINAL_BASE_URL"),
        "HTTP_PROXY": os.environ.get("HTTP_PROXY"),
        "HTTPS_PROXY": os.environ.get("HTTPS_PROXY"),
    }

    # Store original target URL in ANTHROPIC_ORIGINAL_BASE_URL if not already set
    # This is used by the proxy to know where to forward requests
    if "ANTHROPIC_ORIGINAL_BASE_URL" not in os.environ:
        target = resolve_target_url_from_env({})  # Check only os.environ
        if target:
            os.environ["ANTHROPIC_ORIGINAL_BASE_URL"] = target
            snapshot["ANTHROPIC_ORIGINAL_BASE_URL"] = None  # Was not set

    # Set proxy URL
    os.environ["ANTHROPIC_BASE_URL"] = proxy_url

    # Remove HTTP_PROXY and HTTPS_PROXY (our proxy will forward to them)
    for proxy_var in ["HTTP_PROXY", "HTTPS_PROXY"]:
        os.environ.pop(proxy_var, None)

    # Handle Foundry-specific env vars
    if is_truthy_env(os.environ.get(FOUNDRY_USE_ENV)):
        snapshot[FOUNDRY_BASE_URL_ENV] = os.environ.get(FOUNDRY_BASE_URL_ENV)
        snapshot[FOUNDRY_RESOURCE_ENV] = os.environ.get(FOUNDRY_RESOURCE_ENV)

        os.environ[FOUNDRY_BASE_URL_ENV] = proxy_url
        os.environ.pop(FOUNDRY_RESOURCE_ENV, None)

    # Handle Bedrock-specific env vars
    if is_truthy_env(os.environ.get(BEDROCK_USE_ENV)):
        snapshot[BEDROCK_BASE_URL_ENV] = os.environ.get(BEDROCK_BASE_URL_ENV)
        os.environ[BEDROCK_BASE_URL_ENV] = proxy_url

    return snapshot



"""
Rollout wrapper for Anthropic chat completions instrumentation.

This module adds caching and override capabilities to Anthropic chat completions
instrumentation during rollout sessions.
"""

import json
from typing import Any, AsyncGenerator, Generator, Optional

from opentelemetry.trace import Span

from anthropic.types import (
    Message,
    RawMessageStreamEvent,
    InputJSONDelta,
    MessageDeltaUsage,
    RawContentBlockDeltaEvent,
    RawContentBlockStartEvent,
    RawContentBlockStopEvent,
    RawMessageDeltaEvent,
    RawMessageStartEvent,
    RawMessageStopEvent,
    TextDelta,
    ThinkingDelta,
    Usage,
)

from anthropic.types.raw_message_delta_event import Delta
from lmnr.sdk.log import get_default_logger
from lmnr.sdk.rollout.instrumentation import RolloutInstrumentationWrapper
from lmnr.sdk.rollout_control import is_rollout_mode

logger = get_default_logger(__name__)


class AnthropicRolloutWrapper(RolloutInstrumentationWrapper):
    """
    Rollout wrapper specific to Anthropic chat completions instrumentation.

    Handles:
    - Converting cached responses to Anthropic Message format
    - Applying overrides to system prompts and tool definitions
    - Setting rollout-specific span attributes
    """

    def apply_anthropic_overrides(
        self, kwargs: dict[str, Any], path: str, span: Span | None = None
    ) -> dict[str, Any]:
        overrides = self.get_overrides(path)
        if not overrides:
            return kwargs

        modified_kwargs = kwargs.copy()

        if "system" in overrides:
            system_override = overrides["system"]
            logger.debug(f"Applying system override for {path}")

            if span and span.is_recording():
                try:
                    # For Anthropic, we store system prompt separately in gen_ai.input.messages if present
                    # or it might be passed as a 'system' parameter in kwargs.
                    # We update the span attribute to reflect the override.
                    if input_messages_raw := span.attributes.get(
                        "gen_ai.input.messages"
                    ):
                        input_messages = json.loads(input_messages_raw)
                        if input_messages and input_messages[0].get("role") == "system":
                            input_messages[0]["content"] = system_override
                        else:
                            input_messages.insert(
                                0, {"role": "system", "content": system_override}
                            )

                        span.set_attribute(
                            "gen_ai.input.messages",
                            json.dumps(input_messages),
                        )
                except Exception:
                    pass

            modified_kwargs["system"] = system_override

        if "tools" in overrides:
            # Anthropic tools are passed as a list of dicts
            modified_kwargs["tools"] = overrides["tools"]

        return modified_kwargs

    def cached_response_to_anthropic(
        self, cached_span: dict[str, Any]
    ) -> Message | None:
        """
        Convert cached span data to Anthropic Message response.
        Tries raw response first, then falls back to reconstruction.
        """
        attributes = cached_span.get("attributes", {})
        if raw_response := attributes.get("lmnr.sdk.raw.response"):
            try:
                response_dict = (
                    raw_response
                    if isinstance(raw_response, dict)
                    else json.loads(raw_response)
                )
                if response_dict:
                    # Reset usage to 0 tokens/cost for cached responses
                    if "usage" in response_dict:
                        response_dict["usage"] = {
                            "input_tokens": 0,
                            "output_tokens": 0,
                        }
                    return Message.model_validate(response_dict)
            except Exception as e:
                logger.debug(f"Failed to parse raw response: {e}")

        try:
            output_str = attributes.get("gen_ai.output.messages")
            if not output_str:
                logger.warning("Cached span has no output messages")
                return None

            output = json.loads(output_str)
            if not isinstance(output, list) or not output:
                return None

            # Anthropic instrumentation saves output as a list of choices, usually just one
            first_choice = output[0]
            content_blocks = first_choice.get("content", [])

            return Message(
                id=attributes.get("gen_ai.response.id", "cached"),
                model=attributes.get("gen_ai.response.model", "unknown"),
                content=content_blocks,
                role="assistant",
                stop_reason=first_choice.get("stop_reason", "end_turn"),
                type="message",
                usage=Usage(input_tokens=0, output_tokens=0),
            )

        except Exception as e:
            logger.debug(
                f"Failed to convert cached response to Anthropic format: {e}",
                exc_info=True,
            )
            return None

    def wrap_create(
        self,
        wrapped,
        instance,
        args,
        kwargs,
        span: Span | None = None,
        is_streaming: bool = False,
        is_async: bool = False,
    ) -> Any:
        if not self.should_use_rollout():
            return wrapped(*args, **kwargs)

        span_path = self._get_span_path_from_span(span) if span else None
        if not span_path:
            return wrapped(*args, **kwargs)

        current_index = self.get_current_index_for_path(span_path)
        logger.debug(f"Anthropic create call at {span_path}:{current_index}")

        if self.should_use_cache(span_path, current_index):
            cached_span = self.get_cached_response(span_path, current_index)
            if cached_span:
                logger.debug(
                    f"Using cached response for Anthropic at {span_path}:{current_index}"
                )
                response = self.cached_response_to_anthropic(cached_span)
                if response:
                    try:
                        if span and span.is_recording():
                            span.set_attributes(
                                {
                                    "lmnr.span.type": "CACHED",
                                    "lmnr.rollout.cache_index": current_index,
                                    "lmnr.span.original_type": "LLM",
                                }
                            )
                    except Exception:
                        pass

                    if is_streaming:
                        if is_async:
                            return self._as_coroutine(
                                self._create_async_cached_stream(response)
                            )
                        else:
                            return self._create_cached_stream(response)

                    if is_async:
                        return self._as_coroutine(response)
                    return response

        modified_kwargs = self.apply_anthropic_overrides(kwargs, span_path, span)
        logger.debug(f"Executing live Anthropic call for {span_path}:{current_index}")
        return wrapped(*args, **modified_kwargs)

    def _create_cached_stream(
        self, response: Message
    ) -> Generator[RawMessageStreamEvent, None, None]:
        """Yield a cached response as a sequence of streaming events."""

        # 1. Message Start
        yield RawMessageStartEvent(
            message=Message(
                id=response.id,
                model=response.model,
                role=response.role,
                content=[],  # Content is empty in message_start
                type="message",
                usage=Usage(input_tokens=0, output_tokens=0),
            ),
            type="message_start",
        )

        # 2. Content Blocks
        for i, block in enumerate(response.content):
            # Content Block Start
            yield RawContentBlockStartEvent(
                content_block=block,
                index=i,
                type="content_block_start",
            )

            # Content Block Delta (send entire content as one delta)
            delta_obj = None
            if block.type == "text":
                delta_obj = TextDelta(text=block.text, type="text_delta")
            elif block.type == "tool_use":
                # For tool use, we send the whole input as one delta
                input_json = (
                    json.dumps(block.input)
                    if not isinstance(block.input, str)
                    else block.input
                )
                delta_obj = InputJSONDelta(
                    partial_json=input_json, type="input_json_delta"
                )
            elif block.type == "thinking":
                delta_obj = ThinkingDelta(
                    thinking=block.thinking, type="thinking_delta"
                )

            if delta_obj:
                yield RawContentBlockDeltaEvent(
                    delta=delta_obj,
                    index=i,
                    type="content_block_delta",
                )

            # Content Block Stop
            yield RawContentBlockStopEvent(
                index=i,
                type="content_block_stop",
            )

        # 3. Message Delta
        yield RawMessageDeltaEvent(
            delta=Delta(
                stop_reason=response.stop_reason,
                stop_sequence=response.stop_sequence,
            ),
            type="message_delta",
            usage=MessageDeltaUsage(output_tokens=0),
        )

        # 4. Message Stop
        yield RawMessageStopEvent(type="message_stop")

    async def _create_async_cached_stream(
        self, response: Message
    ) -> AsyncGenerator[RawMessageStreamEvent, None]:
        """Yield a cached response as a sequence of async streaming events."""
        for event in self._create_cached_stream(response):
            yield event

    async def _as_coroutine(self, response: Any) -> Any:
        """Return a response as a coroutine."""
        return response

    def _get_span_path_from_span(self, span: Span) -> Optional[str]:
        """Get the span path from the span's attributes."""
        try:
            path_list = span.attributes.get("lmnr.span.path")
            if path_list:
                return ".".join(path_list)
        except Exception:
            pass
        return None


_anthropic_rollout_wrapper: AnthropicRolloutWrapper | None = None


def get_anthropic_rollout_wrapper() -> AnthropicRolloutWrapper | None:
    global _anthropic_rollout_wrapper

    if not is_rollout_mode():
        return None

    if _anthropic_rollout_wrapper is None:
        try:
            _anthropic_rollout_wrapper = AnthropicRolloutWrapper()
        except Exception as e:
            logger.error(f"Failed to create Anthropic rollout wrapper: {e}")
            return None

    return _anthropic_rollout_wrapper



import json
import logging

from opentelemetry.semconv._incubating.attributes.gen_ai_attributes import (
    GEN_AI_REQUEST_FREQUENCY_PENALTY,
    GEN_AI_REQUEST_MAX_TOKENS,
    GEN_AI_REQUEST_MODEL,
    GEN_AI_REQUEST_PRESENCE_PENALTY,
    GEN_AI_REQUEST_TEMPERATURE,
    GEN_AI_REQUEST_TOP_P,
    GEN_AI_RESPONSE_ID,
    GEN_AI_RESPONSE_MODEL,
    GEN_AI_USAGE_INPUT_TOKENS,
    GEN_AI_USAGE_OUTPUT_TOKENS,
)

from lmnr.sdk.utils import json_dumps

from .event_models import AnthropicResponseMessage
from .utils import (
    _aextract_response_data,
    _extract_response_data,
    dont_throw,
    model_as_dict,
    set_span_attribute,
    should_send_prompts,
)

logger = logging.getLogger(__name__)


def _process_content_for_message(content):
    """Convert message content to a serializable format.

    For strings, returns as-is. For lists, converts each item via model_as_dict.
    """
    if isinstance(content, str):
        return content
    elif isinstance(content, list):
        return [model_as_dict(item) for item in content]
    return content


@dont_throw
async def aset_input_attributes(span, kwargs):
    from .utils import set_span_attribute

    set_span_attribute(span, GEN_AI_REQUEST_MODEL, kwargs.get("model"))
    set_span_attribute(
        span, GEN_AI_REQUEST_MAX_TOKENS, kwargs.get("max_tokens_to_sample")
    )
    set_span_attribute(span, GEN_AI_REQUEST_TEMPERATURE, kwargs.get("temperature"))
    set_span_attribute(span, GEN_AI_REQUEST_TOP_P, kwargs.get("top_p"))
    set_span_attribute(
        span, GEN_AI_REQUEST_FREQUENCY_PENALTY, kwargs.get("frequency_penalty")
    )
    set_span_attribute(
        span, GEN_AI_REQUEST_PRESENCE_PENALTY, kwargs.get("presence_penalty")
    )
    set_span_attribute(span, "llm.is_streaming", kwargs.get("stream"))
    set_span_attribute(
        span, "anthropic.request.service_tier", kwargs.get("service_tier")
    )

    if should_send_prompts():
        if kwargs.get("prompt") is not None:
            # Legacy completions API
            messages = [{"role": "user", "content": kwargs.get("prompt")}]
            set_span_attribute(span, "gen_ai.input.messages", json_dumps(messages))

        elif kwargs.get("messages") is not None:
            messages = []

            # Add system message if present
            if kwargs.get("system"):
                system_content = _process_content_for_message(kwargs.get("system"))
                messages.append({"role": "system", "content": system_content})

            # Add all user/assistant messages
            for i, message in enumerate(kwargs.get("messages")):
                msg = dict(message)
                content = msg.get("content")
                if content is not None:
                    msg["content"] = _process_content_for_message(content)
                messages.append(msg)

            set_span_attribute(span, "gen_ai.input.messages", json_dumps(messages))

        if kwargs.get("tools") is not None:
            set_span_attribute(
                span,
                "gen_ai.tool.definitions",
                json_dumps(kwargs.get("tools")),
            )


def _build_output_from_response(response):
    """Build the output messages list from a non-streaming Anthropic response.

    Returns a list of content block dicts as returned by the Anthropic API,
    representing the single candidate/choice.
    """
    result = {
        "role": response.get("role", "assistant"),
        "content": [],
    }

    stop_reason = response.get("stop_reason")
    if stop_reason:
        result["stop_reason"] = stop_reason

    if response.get("completion"):
        result["content"].append(
            {
                "type": "text",
                "text": response.get("completion"),
            }
        )
    elif response.get("content"):
        for block in response.get("content"):
            result["content"].append(model_as_dict(block))

    return [result]


async def _aset_span_completions(span, response):
    if not should_send_prompts():
        return

    response = await _aextract_response_data(response)
    output = _build_output_from_response(response)
    set_span_attribute(span, "gen_ai.output.messages", json_dumps(output))


def _set_span_completions(span, response):
    if not should_send_prompts():
        return
    from .utils import set_span_attribute

    output = _build_output_from_response(response)
    set_span_attribute(span, "gen_ai.output.messages", json_dumps(output))


@dont_throw
async def aset_response_attributes(span, response):
    response = await _aextract_response_data(response)
    set_span_attribute(span, GEN_AI_RESPONSE_MODEL, response.get("model"))
    set_span_attribute(span, GEN_AI_RESPONSE_ID, response.get("id"))

    if usage := response.get("usage"):
        if hasattr(usage, "service_tier"):
            set_span_attribute(
                span,
                "anthropic.response.service_tier",
                usage.service_tier,
            )
        prompt_tokens = usage.input_tokens
        completion_tokens = usage.output_tokens
        set_span_attribute(span, GEN_AI_USAGE_INPUT_TOKENS, prompt_tokens)
        set_span_attribute(span, GEN_AI_USAGE_OUTPUT_TOKENS, completion_tokens)

    await _aset_span_completions(span, response)


@dont_throw
def set_response_attributes(span, response):
    response = _extract_response_data(response)
    set_span_attribute(span, GEN_AI_RESPONSE_MODEL, response.get("model"))
    set_span_attribute(span, GEN_AI_RESPONSE_ID, response.get("id"))

    if usage := response.get("usage"):
        if hasattr(usage, "service_tier"):
            set_span_attribute(
                span,
                "anthropic.response.service_tier",
                usage.service_tier,
            )
        prompt_tokens = usage.input_tokens
        completion_tokens = usage.output_tokens
        set_span_attribute(span, GEN_AI_USAGE_INPUT_TOKENS, prompt_tokens)
        set_span_attribute(span, GEN_AI_USAGE_OUTPUT_TOKENS, completion_tokens)

    _set_span_completions(span, response)


@dont_throw
def set_streaming_response_attributes(
    span, complete_response_events
) -> AnthropicResponseMessage | None:
    if not span or not span.is_recording() or not complete_response_events:
        return None

    # Build an output in the same format as non-streaming responses
    result = {
        "role": "assistant",
        "content": [],
    }

    finish_reason = None
    for event in complete_response_events:
        event_type = event.get("type")
        if event.get("finish_reason"):
            finish_reason = event.get("finish_reason")

        if event_type == "thinking":
            result["content"].append(
                {
                    "type": "thinking",
                    "thinking": event.get("text", ""),
                }
            )
        elif event_type == "tool_use":
            tool_input = event.get("input", "")
            # input may be a stringified JSON from streaming, try to parse
            if isinstance(tool_input, str):
                try:
                    tool_input = json.loads(tool_input)
                except (json.JSONDecodeError, TypeError):
                    pass
            result["content"].append(
                {
                    "type": "tool_use",
                    "id": event.get("id", ""),
                    "name": event.get("name", ""),
                    "input": tool_input,
                }
            )
        else:
            # text block
            text = event.get("text", "")
            if text:
                result["content"].append(
                    {
                        "type": "text",
                        "text": text,
                    }
                )

    if finish_reason:
        result["stop_reason"] = finish_reason

    if should_send_prompts():
        set_span_attribute(span, "gen_ai.output.messages", json_dumps([result]))

    return result



from opentelemetry.trace import Span
from lmnr.opentelemetry_lib.tracing.span import LaminarSpan
from lmnr.opentelemetry_lib.tracing.attributes import (
    ASSOCIATION_PROPERTIES,
    ROLLOUT_SESSION_ID,
    USER_ID,
    SESSION_ID,
    TRACE_TYPE,
)
from lmnr.opentelemetry_lib.tracing.context import (
    get_current_context,
    attach_context,
    set_value,
    CONTEXT_USER_ID_KEY,
    CONTEXT_SESSION_ID_KEY,
    CONTEXT_ROLLOUT_SESSION_ID_KEY,
    CONTEXT_TRACE_TYPE_KEY,
    CONTEXT_METADATA_KEY,
)
from lmnr.sdk.log import get_default_logger

logger = get_default_logger(__name__)


def set_association_props_in_context(span: Span):
    """Set association properties from span in context before push_span_context.

    Returns the token that needs to be detached when the span ends.
    """
    if not isinstance(span, LaminarSpan):
        return None

    try:
        props = span.laminar_association_properties
        user_id_key = f"{ASSOCIATION_PROPERTIES}.{USER_ID}"
        session_id_key = f"{ASSOCIATION_PROPERTIES}.{SESSION_ID}"
        rollout_session_id_key = f"{ASSOCIATION_PROPERTIES}.{ROLLOUT_SESSION_ID}"
        trace_type_key = f"{ASSOCIATION_PROPERTIES}.{TRACE_TYPE}"

        # Extract values from props
        extracted_user_id = props.get(user_id_key)
        extracted_session_id = props.get(session_id_key)
        extracted_rollout_session_id = props.get(rollout_session_id_key)
        extracted_trace_type = props.get(trace_type_key)

        # Extract metadata from props (keys without ASSOCIATION_PROPERTIES prefix)
        metadata_dict = {}
        for key, value in props.items():
            if not key.startswith(f"{ASSOCIATION_PROPERTIES}."):
                metadata_dict[key] = value

        # Set context with association props
        current_ctx = get_current_context()
        ctx_with_props = current_ctx
        if extracted_user_id:
            ctx_with_props = set_value(
                CONTEXT_USER_ID_KEY, extracted_user_id, ctx_with_props
            )
        if extracted_session_id:
            ctx_with_props = set_value(
                CONTEXT_SESSION_ID_KEY, extracted_session_id, ctx_with_props
            )
        if extracted_rollout_session_id:
            ctx_with_props = set_value(
                CONTEXT_ROLLOUT_SESSION_ID_KEY,
                extracted_rollout_session_id,
                ctx_with_props,
            )
        if extracted_trace_type:
            ctx_with_props = set_value(
                CONTEXT_TRACE_TYPE_KEY, extracted_trace_type, ctx_with_props
            )
        if metadata_dict:
            ctx_with_props = set_value(
                CONTEXT_METADATA_KEY, metadata_dict, ctx_with_props
            )

        return attach_context(ctx_with_props)

    except Exception:
        logger.warning(
            "[set_association_props_in_context] failed to set association props in context",
            exc_info=True,
        )
        return None



import logging
import time

from opentelemetry import context as context_api
from opentelemetry.trace import Span, Tracer
from ..shared import (
    _set_span_attribute,
    model_as_dict,
)
from ..utils import (
    _with_tracer_wrapper,
    dont_throw,
)
from lmnr.opentelemetry_lib.tracing.context import (
    get_current_context,
    get_event_attributes_from_context,
)
from opentelemetry.instrumentation.utils import _SUPPRESS_INSTRUMENTATION_KEY
from opentelemetry.semconv.attributes.error_attributes import ERROR_TYPE
from opentelemetry.semconv._incubating.attributes.gen_ai_attributes import (
    GEN_AI_REQUEST_MODEL,
    GEN_AI_RESPONSE_MODEL,
    GEN_AI_USAGE_INPUT_TOKENS,
    GEN_AI_USAGE_OUTPUT_TOKENS,
)
from opentelemetry.trace import Status, StatusCode

from openai._legacy_response import LegacyAPIResponse
from openai.types.beta.threads.run import Run

logger = logging.getLogger(__name__)

assistants = {}
runs = {}


# We are not reusing the safe_start_span from shared utils because we need to pass the
# start_time parameter, which we don't want to expose on Laminar.start_span.
# Assistants API is deprecated anyway, so it's not risky to leave old `tracer` path here.
def _safe_start_span(
    tracer: Tracer,
    name: str,
    start_time: int | None = None,
) -> Span | None:
    try:
        return tracer.start_span(
            name,
            start_time=start_time,
            context=get_current_context(),
            attributes={
                "gen_ai.system": "openai",
                "lmnr.span.type": "LLM",
            },
        )
    except Exception:
        logger.debug(f"[openai assistants] Failed to start span: {name}", exc_info=True)
        return None


@_with_tracer_wrapper
def assistants_create_wrapper(tracer, wrapped, instance, args, kwargs):
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
        return wrapped(*args, **kwargs)

    response = wrapped(*args, **kwargs)

    assistants[response.id] = {
        "model": kwargs.get("model"),
        "instructions": kwargs.get("instructions"),
    }

    return response


@_with_tracer_wrapper
def runs_create_wrapper(tracer, wrapped, instance, args, kwargs):
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
        return wrapped(*args, **kwargs)

    thread_id = kwargs.get("thread_id")
    instructions = kwargs.get("instructions")

    try:
        response = wrapped(*args, **kwargs)
        response_dict = model_as_dict(response)

        runs[thread_id] = {
            "start_time": time.time_ns(),
            "assistant_id": kwargs.get("assistant_id"),
            "instructions": instructions,
            "run_id": response_dict.get("id"),
        }

        return response
    except Exception as e:
        runs[thread_id] = {
            "exception": e,
            "end_time": time.time_ns(),
        }
        raise


@_with_tracer_wrapper
def runs_retrieve_wrapper(tracer, wrapped, instance, args, kwargs):
    @dont_throw
    def process_response(response):
        if type(response) is LegacyAPIResponse:
            parsed_response = response.parse()
        else:
            parsed_response = response
        assert type(parsed_response) is Run

        if parsed_response.thread_id in runs:
            thread_id = parsed_response.thread_id
            runs[thread_id]["end_time"] = time.time_ns()
            if parsed_response.usage:
                runs[thread_id]["usage"] = parsed_response.usage

    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
        return wrapped(*args, **kwargs)

    try:
        response = wrapped(*args, **kwargs)
        process_response(response)
        return response
    except Exception as e:
        thread_id = kwargs.get("thread_id")
        if thread_id in runs:
            runs[thread_id]["exception"] = e
            runs[thread_id]["end_time"] = time.time_ns()
        raise


@_with_tracer_wrapper
def messages_list_wrapper(tracer, wrapped, instance, args, kwargs):
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
        return wrapped(*args, **kwargs)

    id = kwargs.get("thread_id")

    response = wrapped(*args, **kwargs)

    response_dict = model_as_dict(response)
    if id not in runs:
        return response

    run = runs[id]
    messages = sorted(
        response_dict.get("data", []), key=lambda x: x.get("created_at", 0)
    )

    span = _safe_start_span(
        tracer,
        "openai.assistant.run",
        start_time=run.get("start_time"),
    )
    if span is None:
        return response

    if exception := run.get("exception"):
        span.set_attribute(ERROR_TYPE, exception.__class__.__name__)
        span.record_exception(exception, attributes=get_event_attributes_from_context())
        span.set_status(Status(StatusCode.ERROR, str(exception)))
        span.end(run.get("end_time"))

    prompt_index = 0
    if assistants.get(run["assistant_id"]) is not None:
        assistant = assistants[run["assistant_id"]]

        _set_span_attribute(
            span,
            "gen_ai.system",
            "openai",
        )
        _set_span_attribute(
            span,
            GEN_AI_REQUEST_MODEL,
            assistant["model"],
        )
        _set_span_attribute(
            span,
            GEN_AI_RESPONSE_MODEL,
            assistant["model"],
        )
        _set_span_attribute(span, f"gen_ai.prompt.{prompt_index}.role", "system")
        _set_span_attribute(
            span,
            f"gen_ai.prompt.{prompt_index}.content",
            assistant["instructions"],
        )
        prompt_index += 1
    _set_span_attribute(span, f"gen_ai.prompt.{prompt_index}.role", "system")
    _set_span_attribute(
        span,
        f"gen_ai.prompt.{prompt_index}.content",
        run["instructions"],
    )
    prompt_index += 1

    completion_index = 0
    for msg in messages:
        prefix = f"gen_ai.completion.{completion_index}"
        content = msg.get("content")

        message_content = content[0].get("text").get("value")
        message_role = msg.get("role")
        if message_role in ["user", "system"]:
            _set_span_attribute(
                span,
                f"gen_ai.prompt.{prompt_index}.role",
                message_role,
            )
            _set_span_attribute(
                span,
                f"gen_ai.prompt.{prompt_index}.content",
                message_content,
            )
            prompt_index += 1
        else:

            _set_span_attribute(span, f"{prefix}.role", msg.get("role"))
            _set_span_attribute(span, f"{prefix}.content", message_content)
            _set_span_attribute(
                span, f"gen_ai.response.{completion_index}.id", msg.get("id")
            )
            completion_index += 1

    if run.get("usage"):
        usage_dict = model_as_dict(run.get("usage"))
        _set_span_attribute(
            span,
            GEN_AI_USAGE_OUTPUT_TOKENS,
            usage_dict.get("completion_tokens"),
        )
        _set_span_attribute(
            span,
            GEN_AI_USAGE_INPUT_TOKENS,
            usage_dict.get("prompt_tokens"),
        )

    span.end(run.get("end_time"))

    return response


@_with_tracer_wrapper
def runs_create_and_stream_wrapper(tracer, wrapped, instance, args, kwargs):
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
        return wrapped(*args, **kwargs)

    assistant_id = kwargs.get("assistant_id")
    instructions = kwargs.get("instructions")

    span = _safe_start_span(
        tracer,
        "openai.assistant.run_stream",
    )
    if span is None:
        return wrapped(*args, **kwargs)

    i = 0
    if assistants.get(assistant_id) is not None:
        _set_span_attribute(
            span, GEN_AI_REQUEST_MODEL, assistants[assistant_id]["model"]
        )
        _set_span_attribute(
            span,
            "gen_ai.system",
            "openai",
        )
        _set_span_attribute(
            span,
            GEN_AI_RESPONSE_MODEL,
            assistants[assistant_id]["model"],
        )
        _set_span_attribute(span, f"gen_ai.prompt.{i}.role", "system")
        _set_span_attribute(
            span,
            f"gen_ai.prompt.{i}.content",
            assistants[assistant_id]["instructions"],
        )
        i += 1
    _set_span_attribute(span, f"gen_ai.prompt.{i}.role", "system")
    _set_span_attribute(span, f"gen_ai.prompt.{i}.content", instructions)

    from ..v1.event_handler_wrapper import (
        EventHandlerWrapper,
    )

    kwargs["event_handler"] = EventHandlerWrapper(
        original_handler=kwargs["event_handler"],
        span=span,
    )

    try:
        response = wrapped(*args, **kwargs)
        return response
    except Exception as e:
        span.set_attribute(ERROR_TYPE, e.__class__.__name__)
        span.record_exception(e, attributes=get_event_attributes_from_context())
        span.set_status(Status(StatusCode.ERROR, str(e)))
        span.end()
        raise



from typing import Collection

from opentelemetry._events import get_event_logger
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from ..shared.chat_wrappers import (
    achat_wrapper,
    chat_wrapper,
)
from ..shared.completion_wrappers import (
    acompletion_wrapper,
    completion_wrapper,
)
from ..shared.config import Config
from ..shared.embeddings_wrappers import (
    aembeddings_wrapper,
    embeddings_wrapper,
)
from ..version import __version__
from opentelemetry.instrumentation.utils import unwrap
from opentelemetry.trace import get_tracer
from wrapt import wrap_function_wrapper

_instruments = ("openai >= 0.27.0", "openai < 1.0.0")


class OpenAIV0Instrumentor(BaseInstrumentor):
    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        tracer_provider = kwargs.get("tracer_provider")
        tracer = get_tracer(__name__, __version__, tracer_provider)

        if not Config.use_legacy_attributes:
            event_logger_provider = kwargs.get("event_logger_provider")
            Config.event_logger = get_event_logger(
                __name__, __version__, event_logger_provider=event_logger_provider
            )

        wrap_function_wrapper(
            "openai",
            "Completion.create",
            completion_wrapper(tracer),
        )

        wrap_function_wrapper(
            "openai",
            "Completion.acreate",
            acompletion_wrapper(tracer),
        )
        wrap_function_wrapper(
            "openai",
            "ChatCompletion.create",
            chat_wrapper(
                tracer,
            ),
        )
        wrap_function_wrapper(
            "openai",
            "ChatCompletion.acreate",
            achat_wrapper(
                tracer,
            ),
        )
        wrap_function_wrapper(
            "openai",
            "Embedding.create",
            embeddings_wrapper(
                tracer,
            ),
        )
        wrap_function_wrapper(
            "openai",
            "Embedding.acreate",
            aembeddings_wrapper(
                tracer,
            ),
        )

    def _uninstrument(self, **kwargs):
        unwrap("openai", "Completion.create")
        unwrap("openai", "Completion.acreate")
        unwrap("openai", "ChatCompletion.create")
        unwrap("openai", "ChatCompletion.acreate")
        unwrap("openai", "Embedding.create")
        unwrap("openai", "Embedding.acreate")



"""
Rollout wrapper for OpenAI chat completions instrumentation.

This module adds caching and override capabilities to OpenAI chat completions
instrumentation during rollout sessions.
"""

import json
from typing import Any, AsyncGenerator, Generator

from openai.types.chat import ChatCompletion, ChatCompletionChunk
from openai.types.chat.chat_completion import Choice
from openai.types.chat.chat_completion_chunk import (
    Choice as ChunkChoice,
    ChoiceDelta,
)
from openai.types.chat.chat_completion_message import (
    ChatCompletionMessage,
    FunctionCall,
)
from openai.types.chat.chat_completion_message_function_tool_call import (
    ChatCompletionMessageFunctionToolCall,
    Function as ToolCallFunction,
)
from opentelemetry.trace import Span

from lmnr.sdk.log import get_default_logger
from lmnr.sdk.rollout.instrumentation import RolloutInstrumentationWrapper
from lmnr.sdk.rollout_control import is_rollout_mode

logger = get_default_logger(__name__)


class OpenAIRolloutWrapper(RolloutInstrumentationWrapper):
    """
    Rollout wrapper specific to OpenAI chat completions instrumentation.

    Handles:
    - Converting cached responses to OpenAI ChatCompletion format
    - Applying overrides to system prompts and tool definitions
    - Setting rollout-specific span attributes
    """

    def apply_openai_overrides(
        self, kwargs: dict[str, Any], path: str, span: Span | None = None
    ) -> dict[str, Any]:
        overrides = self.get_overrides(path)
        if not overrides:
            return kwargs

        modified_kwargs = kwargs.copy()

        if "system" in overrides:
            system_override = overrides["system"]
            logger.debug(f"Applying system override for {path}")

            if span and span.is_recording():
                try:
                    if input_messages_raw := span.attributes.get(
                        "gen_ai.input.messages"
                    ):
                        input_messages = json.loads(input_messages_raw)
                        if input_messages and input_messages[0].get("role") == "system":
                            input_messages[0]["content"] = system_override
                            span.set_attribute(
                                "gen_ai.input.messages",
                                json.dumps(input_messages),
                            )
                except Exception:
                    pass

            messages = modified_kwargs.get("messages", [])
            if messages and isinstance(messages, list):
                modified_kwargs["messages"] = list(messages)
                if modified_kwargs["messages"][0].get("role") == "system":
                    modified_kwargs["messages"][0] = {
                        **modified_kwargs["messages"][0],
                        "content": system_override,
                    }
                else:
                    modified_kwargs["messages"].insert(
                        0, {"role": "system", "content": system_override}
                    )

        if "tools" in overrides:
            tool_overrides = overrides["tools"]
            logger.debug(
                f"Applying tool overrides for {path}: {len(tool_overrides)} tools"
            )

            existing_tools = modified_kwargs.get("tools", [])
            updated_tools = self._apply_tool_overrides(existing_tools, tool_overrides)
            if updated_tools is not None:
                modified_kwargs["tools"] = updated_tools

        return modified_kwargs

    def _apply_tool_overrides(
        self,
        existing_tools: list[dict[str, Any]] | None,
        tool_overrides: list[dict[str, Any]],
    ) -> list[dict[str, Any]] | None:
        if not tool_overrides:
            return existing_tools

        tools_by_name: dict[str, dict[str, Any]] = {}
        if existing_tools:
            for tool in existing_tools:
                func = tool.get("function", {})
                if name := func.get("name"):
                    tools_by_name[name] = tool

        for override in tool_overrides:
            name = override.get("name")
            if not name:
                continue

            if name in tools_by_name:
                existing_func = tools_by_name[name].get("function", {})
                tools_by_name[name] = {
                    "type": "function",
                    "function": {
                        "name": name,
                        "description": override.get(
                            "description", existing_func.get("description")
                        ),
                        "parameters": override.get(
                            "parameters", existing_func.get("parameters")
                        ),
                    },
                }
            elif "parameters" in override:
                tools_by_name[name] = {
                    "type": "function",
                    "function": {
                        "name": name,
                        "description": override.get("description"),
                        "parameters": override["parameters"],
                    },
                }

        if tools_by_name:
            return list(tools_by_name.values())
        return None

    def cached_response_to_openai(
        self, cached_span: dict[str, Any]
    ) -> ChatCompletion | None:
        """
        Convert cached span data to OpenAI ChatCompletion response.
        Tries raw response first, then falls back to legacy parsing.
        """
        if raw_response := cached_span.get("attributes", {}).get(
            "lmnr.sdk.raw.response"
        ):
            try:
                response_dict = (
                    raw_response
                    if isinstance(raw_response, dict)
                    else json.loads(raw_response)
                )
                if response_dict:
                    return ChatCompletion.model_validate(response_dict)
            except Exception as e:
                logger.debug(f"Failed to parse raw response: {e}")

        try:
            attributes = cached_span.get("attributes", {})
            output_str = cached_span.get("output", "")
            if not output_str:
                logger.warning("Cached span has no output")
                return None

            # NOTE: This is assuming the latest openai instrumentation where we save the output as
            # a list of choices. Not compatible with older versions.
            output = json.loads(output_str)
            if not isinstance(output, list):
                logger.warning(f"Unexpected output format: {type(output)}")
                return None

            choices = []
            for choice_data in output:
                msg = choice_data.get("message", {})

                tool_calls = None
                if raw_tool_calls := msg.get("tool_calls"):
                    tool_calls = [
                        ChatCompletionMessageFunctionToolCall(
                            id=tc.get("id", ""),
                            type="function",
                            function=ToolCallFunction(
                                name=tc.get("function", {}).get("name", ""),
                                arguments=tc.get("function", {}).get("arguments", ""),
                            ),
                        )
                        for tc in raw_tool_calls
                    ]

                function_call = None
                if raw_fc := msg.get("function_call"):
                    function_call = FunctionCall(
                        name=raw_fc.get("name", ""),
                        arguments=raw_fc.get("arguments", ""),
                    )

                choices.append(
                    Choice(
                        index=choice_data.get("index", len(choices)),
                        finish_reason=choice_data.get("finish_reason", "stop"),
                        message=ChatCompletionMessage(
                            role=msg.get("role", "assistant"),
                            content=msg.get("content"),
                            refusal=msg.get("refusal"),
                            tool_calls=tool_calls,
                            function_call=function_call,
                        ),
                    )
                )

            return ChatCompletion(
                id=attributes.get("gen_ai.response.id", "cached"),
                model=attributes.get("gen_ai.response.model", "unknown"),
                choices=choices,
                created=int(cached_span.get("start_time", 0) / 1_000_000_000),
                object="chat.completion",
                usage=None,
            )

        except Exception as e:
            logger.debug(
                f"Failed to convert cached response to OpenAI format: {e}",
                exc_info=True,
            )
            return None

    def _get_span_path_from_span(self, span: Span) -> str | None:
        """Get the span path from the span's attributes (set by the processor)."""
        try:
            path_list = span.attributes.get("lmnr.span.path")
            if path_list:
                return ".".join(path_list)
        except Exception:
            pass
        return None

    def wrap_chat_completion(
        self,
        wrapped,
        instance,
        args,
        kwargs,
        span: Span | None = None,
        is_streaming: bool = False,
        is_async: bool = False,
    ) -> Any:
        if not self.should_use_rollout():
            return wrapped(*args, **kwargs)

        # Read path from the span's attributes rather than Laminar.get_current_span(),
        # because OpenAI spans are created with tracer.start_span() and are not
        # registered in Laminar's context.
        span_path = self._get_span_path_from_span(span) if span else None
        if not span_path:
            return wrapped(*args, **kwargs)

        current_index = self.get_current_index_for_path(span_path)
        logger.debug(f"OpenAI chat completion call at {span_path}:{current_index}")

        if self.should_use_cache(span_path, current_index):
            cached_span = self.get_cached_response(span_path, current_index)
            if cached_span:
                logger.debug(
                    f"Using cached response for OpenAI at {span_path}:{current_index}"
                )
                response = self.cached_response_to_openai(cached_span)
                if response:
                    try:
                        if span and span.is_recording():
                            span.set_attributes(
                                {
                                    "lmnr.span.type": "CACHED",
                                    "lmnr.rollout.cache_index": current_index,
                                    "lmnr.span.original_type": "LLM",
                                }
                            )
                    except Exception:
                        pass

                    if is_streaming:
                        if is_async:
                            return self._create_async_cached_stream(response)
                        else:
                            return self._create_cached_stream(response)
                    return response

        modified_kwargs = self.apply_openai_overrides(kwargs, span_path, span)
        logger.debug(f"Executing live OpenAI call for {span_path}:{current_index}")
        return wrapped(*args, **modified_kwargs)

    def _create_cached_stream(
        self, response: ChatCompletion
    ) -> Generator[ChatCompletionChunk, None, None]:
        """Yield a cached response as a single streaming chunk."""
        response_dict = response.model_dump()

        choices = []
        for choice in response_dict.get("choices", []):
            message = dict(choice.get("message", {}))
            # Streaming tool_calls require an `index` field that the message
            # format doesn't have. Add it before passing to ChoiceDelta.
            if message.get("tool_calls"):
                message["tool_calls"] = [
                    {**tc, "index": i} for i, tc in enumerate(message["tool_calls"])
                ]
            choices.append(
                ChunkChoice(
                    index=choice.get("index", 0),
                    delta=ChoiceDelta(**message),
                    finish_reason=choice.get("finish_reason"),
                )
            )

        yield ChatCompletionChunk(
            id=response_dict.get("id", "cached"),
            model=response_dict.get("model", "unknown"),
            object="chat.completion.chunk",
            created=response_dict.get("created", 0),
            choices=choices,
        )

    async def _create_async_cached_stream(
        self, response: ChatCompletion
    ) -> AsyncGenerator[ChatCompletionChunk, None]:
        """Yield a cached response as a single async streaming chunk."""
        for chunk in self._create_cached_stream(response):
            yield chunk


_openai_rollout_wrapper: OpenAIRolloutWrapper | None = None


def get_openai_rollout_wrapper() -> OpenAIRolloutWrapper | None:
    global _openai_rollout_wrapper

    if not is_rollout_mode():
        return None

    if _openai_rollout_wrapper is None:
        try:
            _openai_rollout_wrapper = OpenAIRolloutWrapper()
        except Exception as e:
            logger.error(f"Failed to create OpenAI debugger wrapper: {e}")
            return None

    return _openai_rollout_wrapper



from typing import Collection

from opentelemetry.instrumentation.instrumentor import BaseInstrumentor

from lmnr.sdk.log import get_default_logger
from ..shared.chat_wrappers import (
    achat_wrapper,
    chat_wrapper,
)
from ..shared.completion_wrappers import (
    acompletion_wrapper,
    completion_wrapper,
)
from ..shared.embeddings_wrappers import (
    aembeddings_wrapper,
    embeddings_wrapper,
)
from .assistant_wrappers import (
    assistants_create_wrapper,
    messages_list_wrapper,
    runs_create_and_stream_wrapper,
    runs_create_wrapper,
    runs_retrieve_wrapper,
)

from .responses_wrappers import (
    async_responses_cancel_wrapper,
    async_responses_get_or_create_wrapper,
    responses_cancel_wrapper,
    responses_get_or_create_wrapper,
)

from ..version import __version__
from opentelemetry.instrumentation.utils import unwrap
from opentelemetry.trace import get_tracer
from wrapt import wrap_function_wrapper


_instruments = ("openai >= 1.0.0",)
logger = get_default_logger(__name__)


class OpenAIV1Instrumentor(BaseInstrumentor):
    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _try_wrap(self, module, function, wrapper):
        """
        Wrap a function if it exists, otherwise do nothing.
        This is useful for handling cases where the function is not available in
        the older versions of the library.

        Args:
            module (str): The module to wrap, e.g. "openai.resources.chat.completions"
            function (str): "Object.function" to wrap, e.g. "Completions.parse"
            wrapper (callable): The wrapper to apply to the function.
        """
        try:
            wrap_function_wrapper(module, function, wrapper)
        except (AttributeError, ModuleNotFoundError, ImportError):
            logger.debug(f"Failed to wrap {module}.{function}")
            pass

    def _instrument(self, **kwargs):
        tracer_provider = kwargs.get("tracer_provider")
        tracer = get_tracer(__name__, __version__, tracer_provider)

        wrap_function_wrapper(
            "openai.resources.chat.completions",
            "Completions.create",
            chat_wrapper(
                tracer,
            ),
        )

        wrap_function_wrapper(
            "openai.resources.completions",
            "Completions.create",
            completion_wrapper(tracer),
        )

        wrap_function_wrapper(
            "openai.resources.embeddings",
            "Embeddings.create",
            embeddings_wrapper(
                tracer,
            ),
        )

        wrap_function_wrapper(
            "openai.resources.chat.completions",
            "AsyncCompletions.create",
            achat_wrapper(
                tracer,
            ),
        )
        wrap_function_wrapper(
            "openai.resources.completions",
            "AsyncCompletions.create",
            acompletion_wrapper(tracer),
        )
        wrap_function_wrapper(
            "openai.resources.embeddings",
            "AsyncEmbeddings.create",
            aembeddings_wrapper(
                tracer,
            ),
        )
        # in newer versions, Completions.parse are out of beta
        self._try_wrap(
            "openai.resources.chat.completions",
            "Completions.parse",
            chat_wrapper(
                tracer,
            ),
        )
        self._try_wrap(
            "openai.resources.chat.completions",
            "AsyncCompletions.parse",
            achat_wrapper(
                tracer,
            ),
        )

        # Beta APIs may not be available consistently in all versions
        self._try_wrap(
            "openai.resources.beta.assistants",
            "Assistants.create",
            assistants_create_wrapper(tracer),
        )
        self._try_wrap(
            "openai.resources.beta.chat.completions",
            "Completions.parse",
            chat_wrapper(
                tracer,
            ),
        )
        self._try_wrap(
            "openai.resources.beta.chat.completions",
            "AsyncCompletions.parse",
            achat_wrapper(
                tracer,
            ),
        )
        self._try_wrap(
            "openai.resources.beta.threads.runs",
            "Runs.create",
            runs_create_wrapper(tracer),
        )
        self._try_wrap(
            "openai.resources.beta.threads.runs",
            "Runs.retrieve",
            runs_retrieve_wrapper(tracer),
        )
        self._try_wrap(
            "openai.resources.beta.threads.runs",
            "Runs.create_and_stream",
            runs_create_and_stream_wrapper(tracer),
        )
        self._try_wrap(
            "openai.resources.beta.threads.messages",
            "Messages.list",
            messages_list_wrapper(tracer),
        )
        self._try_wrap(
            "openai.resources.responses",
            "Responses.create",
            responses_get_or_create_wrapper(tracer),
        )
        self._try_wrap(
            "openai.resources.responses",
            "Responses.retrieve",
            responses_get_or_create_wrapper(tracer),
        )
        self._try_wrap(
            "openai.resources.responses",
            "Responses.cancel",
            responses_cancel_wrapper(tracer),
        )
        self._try_wrap(
            "openai.resources.responses",
            "AsyncResponses.create",
            async_responses_get_or_create_wrapper(tracer),
        )
        self._try_wrap(
            "openai.resources.responses",
            "AsyncResponses.retrieve",
            async_responses_get_or_create_wrapper(tracer),
        )
        self._try_wrap(
            "openai.resources.responses",
            "AsyncResponses.cancel",
            async_responses_cancel_wrapper(tracer),
        )

    def _uninstrument(self, **kwargs):
        self.try_unwrap("openai.resources.chat.completions.Completions", "create")
        self.try_unwrap("openai.resources.completions.Completions", "create")
        self.try_unwrap("openai.resources.embeddings.Embeddings", "create")
        self.try_unwrap("openai.resources.chat.completions.AsyncCompletions", "create")
        self.try_unwrap("openai.resources.completions.AsyncCompletions", "create")
        self.try_unwrap("openai.resources.embeddings.AsyncEmbeddings", "create")
        self.try_unwrap("openai.resources.images.Images", "generate")
        self.try_unwrap("openai.resources.chat.completions.Completions", "parse")
        self.try_unwrap("openai.resources.chat.completions.AsyncCompletions", "parse")
        self.try_unwrap("openai.resources.beta.assistants.Assistants", "create")
        self.try_unwrap("openai.resources.beta.chat.completions.Completions", "parse")
        self.try_unwrap(
            "openai.resources.beta.chat.completions.AsyncCompletions", "parse"
        )
        self.try_unwrap("openai.resources.beta.threads.runs.Runs", "create")
        self.try_unwrap("openai.resources.beta.threads.runs.Runs", "retrieve")
        self.try_unwrap("openai.resources.beta.threads.runs.Runs", "create_and_stream")
        self.try_unwrap("openai.resources.beta.threads.messages.Messages", "list")
        self.try_unwrap("openai.resources.responses.Responses", "create")
        self.try_unwrap("openai.resources.responses.Responses", "retrieve")
        self.try_unwrap("openai.resources.responses.Responses", "cancel")
        self.try_unwrap("openai.resources.responses.AsyncResponses", "create")
        self.try_unwrap("openai.resources.responses.AsyncResponses", "retrieve")
        self.try_unwrap("openai.resources.responses.AsyncResponses", "cancel")

    def try_unwrap(self, module, function):
        try:
            unwrap(module, function)
        except (AttributeError, ModuleNotFoundError, ImportError):
            logger.debug(f"Failed to unwrap {module}.{function}")
            pass



"""OpenTelemetry Anthropic instrumentation"""

import logging
from typing import Callable, Collection

from opentelemetry import context as context_api
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry.instrumentation.utils import _SUPPRESS_INSTRUMENTATION_KEY, unwrap
from opentelemetry.semconv._incubating.attributes.gen_ai_attributes import (
    GEN_AI_USAGE_INPUT_TOKENS,
    GEN_AI_USAGE_OUTPUT_TOKENS,
)
from opentelemetry.trace import Span, Tracer, get_tracer
from opentelemetry.trace.status import Status, StatusCode
from wrapt import wrap_function_wrapper

from anthropic._streaming import AsyncStream, Stream
from lmnr.opentelemetry_lib.opentelemetry.instrumentation.shared.utils import (
    safe_start_span,
)

from .config import Config
from .rollout import get_anthropic_rollout_wrapper
from .span_utils import (
    aset_input_attributes,
    aset_response_attributes,
    set_response_attributes,
)
from .streaming import (
    WrappedAsyncMessageStreamManager,
    WrappedMessageStreamManager,
    abuild_from_streaming_response,
    build_from_streaming_response,
)
from .utils import (
    dont_throw,
    run_async,
    set_span_attribute,
)
from .version import __version__

logger = logging.getLogger(__name__)

_instruments = ("anthropic >= 0.3.11",)


WRAPPED_METHODS = [
    {
        "package": "anthropic.resources.completions",
        "object": "Completions",
        "method": "create",
        "span_name": "anthropic.completion",
    },
    {
        "package": "anthropic.resources.messages",
        "object": "Messages",
        "method": "create",
        "span_name": "anthropic.chat",
    },
    {
        "package": "anthropic.resources.messages",
        "object": "Messages",
        "method": "stream",
        "span_name": "anthropic.chat",
    },
    # This method is on an async resource, but is meant to be called as
    # an async context manager (async with), which we don't need to await;
    # thus, we wrap it with a sync wrapper
    {
        "package": "anthropic.resources.messages",
        "object": "AsyncMessages",
        "method": "stream",
        "span_name": "anthropic.chat",
    },
    # Beta API methods (regular Anthropic SDK)
    {
        "package": "anthropic.resources.beta.messages.messages",
        "object": "Messages",
        "method": "create",
        "span_name": "anthropic.chat",
    },
    {
        "package": "anthropic.resources.beta.messages.messages",
        "object": "Messages",
        "method": "stream",
        "span_name": "anthropic.chat",
    },
    # read note on async with above
    {
        "package": "anthropic.resources.beta.messages.messages",
        "object": "AsyncMessages",
        "method": "stream",
        "span_name": "anthropic.chat",
    },
    # Beta API methods (Bedrock SDK)
    {
        "package": "anthropic.lib.bedrock._beta_messages",
        "object": "Messages",
        "method": "create",
        "span_name": "anthropic.chat",
    },
    {
        "package": "anthropic.lib.bedrock._beta_messages",
        "object": "Messages",
        "method": "stream",
        "span_name": "anthropic.chat",
    },
    # read note on async with above
    {
        "package": "anthropic.lib.bedrock._beta_messages",
        "object": "AsyncMessages",
        "method": "stream",
        "span_name": "anthropic.chat",
    },
]

WRAPPED_AMETHODS = [
    {
        "package": "anthropic.resources.completions",
        "object": "AsyncCompletions",
        "method": "create",
        "span_name": "anthropic.completion",
    },
    {
        "package": "anthropic.resources.messages",
        "object": "AsyncMessages",
        "method": "create",
        "span_name": "anthropic.chat",
    },
    # Beta API async methods (regular Anthropic SDK)
    {
        "package": "anthropic.resources.beta.messages.messages",
        "object": "AsyncMessages",
        "method": "create",
        "span_name": "anthropic.chat",
    },
    # Beta API async methods (Bedrock SDK)
    {
        "package": "anthropic.lib.bedrock._beta_messages",
        "object": "AsyncMessages",
        "method": "create",
        "span_name": "anthropic.chat",
    },
]


def is_streaming_response(response):
    if isinstance(response, (Stream, AsyncStream)):
        return True

    # For cached streams, they are generators, not Message objects.
    # We check for __next__ and __iter__ for sync generators,
    # and __anext__ and __aiter__ for async generators.
    # This prevents identifying Pydantic models (like Message) as streams.
    return (hasattr(response, "__next__") and hasattr(response, "__iter__")) or (
        hasattr(response, "__anext__") and hasattr(response, "__aiter__")
    )


def is_stream_manager(response):
    """Check if response is a MessageStreamManager or AsyncMessageStreamManager"""
    try:
        from anthropic.lib.streaming._messages import (
            AsyncMessageStreamManager,
            MessageStreamManager,
        )

        return isinstance(response, (MessageStreamManager, AsyncMessageStreamManager))
    except ImportError:
        # Check by class name as fallback
        return (
            response.__class__.__name__ == "MessageStreamManager"
            or response.__class__.__name__ == "AsyncMessageStreamManager"
        )


@dont_throw
async def _aset_token_usage(
    span,
    anthropic,
    request,
    response,
):
    # Handle with_raw_response wrapped responses first
    if response and hasattr(response, "parse") and callable(response.parse):
        try:
            response = response.parse()
        except Exception as e:
            logger.debug(f"Failed to parse with_raw_response: {e}")
            return

    usage = getattr(response, "usage", None) if response else None

    if usage:
        prompt_tokens = getattr(usage, "input_tokens", 0)
        cache_read_tokens = getattr(usage, "cache_read_input_tokens", 0) or 0
        cache_creation_tokens = getattr(usage, "cache_creation_input_tokens", 0) or 0
    else:
        prompt_tokens = 0
        cache_read_tokens = 0
        cache_creation_tokens = 0

    input_tokens = prompt_tokens + cache_read_tokens + cache_creation_tokens

    if usage:
        completion_tokens = getattr(usage, "output_tokens", 0)
    else:
        completion_tokens = 0
        if hasattr(anthropic, "count_tokens"):
            completion_attr = getattr(response, "completion", None)
            content_attr = getattr(response, "content", None)
            if completion_attr:
                completion_tokens = await anthropic.count_tokens(completion_attr)
            elif content_attr:
                completion_tokens = await anthropic.count_tokens(content_attr[0].text)

    total_tokens = input_tokens + completion_tokens

    content_attr = getattr(response, "content", None)
    completion_attr = getattr(response, "completion", None)

    set_span_attribute(span, GEN_AI_USAGE_INPUT_TOKENS, input_tokens)
    set_span_attribute(span, GEN_AI_USAGE_OUTPUT_TOKENS, completion_tokens)
    set_span_attribute(span, "llm.usage.total_tokens", total_tokens)

    set_span_attribute(span, "gen_ai.usage.cache_read_input_tokens", cache_read_tokens)
    set_span_attribute(
        span,
        "gen_ai.usage.cache_creation_input_tokens",
        cache_creation_tokens,
    )


@dont_throw
def _set_token_usage(
    span,
    anthropic,
    request,
    response,
):
    # Handle with_raw_response wrapped responses first
    if response and hasattr(response, "parse") and callable(response.parse):
        try:
            response = response.parse()
        except Exception as e:
            logger.debug(f"Failed to parse with_raw_response: {e}")
            return

    usage = getattr(response, "usage", None) if response else None

    if usage:
        prompt_tokens = getattr(usage, "input_tokens", 0)
        cache_read_tokens = getattr(usage, "cache_read_input_tokens", 0) or 0
        cache_creation_tokens = getattr(usage, "cache_creation_input_tokens", 0) or 0
    else:
        prompt_tokens = 0
        cache_read_tokens = 0
        cache_creation_tokens = 0

    input_tokens = prompt_tokens + cache_read_tokens + cache_creation_tokens

    if usage:
        completion_tokens = getattr(usage, "output_tokens", 0)
    else:
        completion_tokens = 0
        if hasattr(anthropic, "count_tokens"):
            completion_attr = getattr(response, "completion", None)
            content_attr = getattr(response, "content", None)
            if completion_attr:
                completion_tokens = anthropic.count_tokens(completion_attr)
            elif content_attr:
                completion_tokens = anthropic.count_tokens(content_attr[0].text)

    total_tokens = input_tokens + completion_tokens

    content_attr = getattr(response, "content", None)
    completion_attr = getattr(response, "completion", None)

    set_span_attribute(span, GEN_AI_USAGE_INPUT_TOKENS, input_tokens)
    set_span_attribute(span, GEN_AI_USAGE_OUTPUT_TOKENS, completion_tokens)
    set_span_attribute(span, "llm.usage.total_tokens", total_tokens)

    set_span_attribute(span, "gen_ai.usage.cache_read_input_tokens", cache_read_tokens)
    set_span_attribute(
        span,
        "gen_ai.usage.cache_creation_input_tokens",
        cache_creation_tokens,
    )


def _with_chat_telemetry_wrapper(func):
    """Helper for providing tracer for wrapper functions. Includes metric collectors."""

    def _with_chat_telemetry(
        tracer,
        to_wrap,
    ):
        def wrapper(wrapped, instance, args, kwargs):
            return func(
                tracer,
                to_wrap,
                wrapped,
                instance,
                args,
                kwargs,
            )

        return wrapper

    return _with_chat_telemetry


@dont_throw
def _handle_input(span: Span, kwargs):
    if not span.is_recording():
        return
    run_async(aset_input_attributes(span, kwargs))


@dont_throw
async def _ahandle_input(span: Span, kwargs):
    if not span.is_recording():
        return
    await aset_input_attributes(span, kwargs)


@dont_throw
def _handle_response(span: Span, response, record_raw_response=False):
    if not span.is_recording():
        return
    set_response_attributes(span, response)

    if record_raw_response:
        try:
            from lmnr.sdk.utils import json_dumps

            from .utils import _extract_response_data, model_as_dict

            response_data = _extract_response_data(response)
            response_dict = model_as_dict(response_data)
            set_span_attribute(span, "lmnr.sdk.raw.response", json_dumps(response_dict))
        except Exception:
            pass


@dont_throw
async def _ahandle_response(span: Span, response, record_raw_response=False):
    if not span.is_recording():
        return
    await aset_response_attributes(span, response)

    if record_raw_response:
        try:
            from lmnr.sdk.utils import json_dumps

            from .utils import _aextract_response_data, model_as_dict

            response_data = await _aextract_response_data(response)
            response_dict = model_as_dict(response_data)
            set_span_attribute(span, "lmnr.sdk.raw.response", json_dumps(response_dict))
        except Exception:
            pass


@_with_chat_telemetry_wrapper
def _wrap(
    tracer: Tracer,
    to_wrap,
    wrapped,
    instance,
    args,
    kwargs,
):
    """Instruments and calls every function defined in TO_WRAP."""
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
        return wrapped(*args, **kwargs)

    span = safe_start_span(
        name=to_wrap.get("span_name"),
        attributes={"gen_ai.system": "anthropic"},
        span_type="LLM",
    )

    if not span:
        logger.warning("Failed to start span for anthropic chat")
        return wrapped(*args, **kwargs)

    _handle_input(span, kwargs)

    rollout_wrapper = get_anthropic_rollout_wrapper()
    is_rollout = rollout_wrapper is not None

    try:
        if rollout_wrapper:
            response = rollout_wrapper.wrap_create(
                wrapped,
                instance,
                args,
                kwargs,
                span=span,
                is_streaming=kwargs.get("stream", False),
                is_async=False,
            )
        else:
            response = wrapped(*args, **kwargs)
    except Exception as e:  # pylint: disable=broad-except
        raise e

    if kwargs.get("stream") or is_streaming_response(response):
        return build_from_streaming_response(
            span,
            response,
            instance._client,
            kwargs,
            record_raw_response=is_rollout,
        )
    elif is_stream_manager(response):
        if response.__class__.__name__ == "AsyncMessageStreamManager":
            return WrappedAsyncMessageStreamManager(
                response,
                span,
                instance._client,
                kwargs,
                record_raw_response=is_rollout,
            )
        else:
            return WrappedMessageStreamManager(
                response,
                span,
                instance._client,
                kwargs,
                record_raw_response=is_rollout,
            )
    elif response:
        try:
            _handle_response(span, response, record_raw_response=is_rollout)
            if span.is_recording():
                _set_token_usage(
                    span,
                    instance._client,
                    kwargs,
                    response,
                )
        except Exception as ex:  # pylint: disable=broad-except
            logger.warning(
                "Failed to set response attributes for anthropic span, error: %s",
                str(ex),
            )

        if span.is_recording():
            span.set_status(Status(StatusCode.OK))
    span.end()
    return response


@_with_chat_telemetry_wrapper
async def _awrap(
    tracer,
    to_wrap,
    wrapped,
    instance,
    args,
    kwargs,
):
    """Instruments and calls every function defined in TO_WRAP."""
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
        return await wrapped(*args, **kwargs)

    span = safe_start_span(
        name=to_wrap.get("span_name"),
        attributes={"gen_ai.system": "anthropic"},
        span_type="LLM",
    )

    if not span:
        logger.warning("Failed to start span for async anthropic chat")
        return await wrapped(*args, **kwargs)

    await _ahandle_input(span, kwargs)

    rollout_wrapper = get_anthropic_rollout_wrapper()
    is_rollout = rollout_wrapper is not None

    try:
        if rollout_wrapper:
            response = await rollout_wrapper.wrap_create(
                wrapped,
                instance,
                args,
                kwargs,
                span=span,
                is_streaming=kwargs.get("stream", False),
                is_async=True,
            )
        else:
            response = await wrapped(*args, **kwargs)
    except Exception as e:  # pylint: disable=broad-except
        raise e

    if kwargs.get("stream") or is_streaming_response(response):
        return abuild_from_streaming_response(
            span,
            response,
            instance._client,
            kwargs,
            record_raw_response=is_rollout,
        )
    elif is_stream_manager(response):
        if response.__class__.__name__ == "AsyncMessageStreamManager":
            return WrappedAsyncMessageStreamManager(
                response,
                span,
                instance._client,
                kwargs,
                record_raw_response=is_rollout,
            )
        else:
            return WrappedMessageStreamManager(
                response,
                span,
                instance._client,
                kwargs,
                record_raw_response=is_rollout,
            )
    elif response:
        await _ahandle_response(span, response, record_raw_response=is_rollout)

        if span.is_recording():
            await _aset_token_usage(
                span,
                instance._client,
                kwargs,
                response,
            )
            span.set_status(Status(StatusCode.OK))
    span.end()
    return response


class AnthropicInstrumentor(BaseInstrumentor):
    """An instrumentor for Anthropic's client library."""

    def __init__(
        self,
        enrich_token_usage: bool = False,
        exception_logger=None,
        use_legacy_attributes: bool = True,
        get_common_metrics_attributes: Callable[[], dict] = lambda: {},
    ):
        super().__init__()
        Config.exception_logger = exception_logger
        Config.enrich_token_usage = enrich_token_usage
        Config.get_common_metrics_attributes = get_common_metrics_attributes
        Config.use_legacy_attributes = use_legacy_attributes

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        tracer_provider = kwargs.get("tracer_provider")
        tracer = get_tracer(__name__, __version__, tracer_provider)

        for wrapped_method in WRAPPED_METHODS:
            wrap_package = wrapped_method.get("package")
            wrap_object = wrapped_method.get("object")
            wrap_method = wrapped_method.get("method")

            try:
                wrap_function_wrapper(
                    wrap_package,
                    f"{wrap_object}.{wrap_method}",
                    _wrap(
                        tracer,
                        wrapped_method,
                    ),
                )
                logger.debug(
                    f"Successfully wrapped {wrap_package}.{wrap_object}.{wrap_method}"
                )
            except Exception as e:
                logger.debug(
                    f"Failed to wrap {wrap_package}.{wrap_object}.{wrap_method}: {e}"
                )
            except ModuleNotFoundError:
                pass  # that's ok, we don't want to fail if some methods do not exist

        for wrapped_method in WRAPPED_AMETHODS:
            wrap_package = wrapped_method.get("package")
            wrap_object = wrapped_method.get("object")
            wrap_method = wrapped_method.get("method")
            try:
                wrap_function_wrapper(
                    wrap_package,
                    f"{wrap_object}.{wrap_method}",
                    _awrap(
                        tracer,
                        wrapped_method,
                    ),
                )
            except Exception:
                pass  # that's ok, we don't want to fail if some methods do not exist

    def _uninstrument(self, **kwargs):
        for wrapped_method in WRAPPED_METHODS:
            wrap_package = wrapped_method.get("package")
            wrap_object = wrapped_method.get("object")
            unwrap(
                f"{wrap_package}.{wrap_object}",
                wrapped_method.get("method"),
            )
        for wrapped_method in WRAPPED_AMETHODS:
            wrap_package = wrapped_method.get("package")
            wrap_object = wrapped_method.get("object")
            unwrap(
                f"{wrap_package}.{wrap_object}",
                wrapped_method.get("method"),
            )



from dataclasses import dataclass
from typing import Any, Literal, TypedDict

from typing_extensions import NotRequired


class AnthropicUsage(TypedDict):
    input_tokens: int
    output_tokens: int
    cache_creation_input_tokens: NotRequired[int]
    cache_read_input_tokens: NotRequired[int]


class AnthropicContentBlock(TypedDict, total=False):
    type: Literal["text", "tool_use", "thinking"]
    text: str
    id: str
    name: str
    input: str | dict[str, Any]
    thinking: str


class AnthropicResponseMessage(TypedDict, total=False):
    id: str
    model: str
    role: Literal["assistant"]
    content: list[AnthropicContentBlock]
    type: Literal["message"]
    usage: AnthropicUsage
    stop_reason: NotRequired[str | None]
    stop_sequence: NotRequired[str | None]


class _FunctionToolCall(TypedDict):
    function_name: str
    arguments: dict[str, Any] | None


class ToolCall(TypedDict):
    """Represents a tool call in the AI model."""

    id: str
    function: _FunctionToolCall
    type: Literal["function"]


class CompletionMessage(TypedDict):
    """Represents a message in the AI model."""

    content: Any
    role: Literal["assistant"]


@dataclass
class MessageEvent:
    """Represents an input event for the AI model."""

    content: Any
    role: str = "user"
    tool_calls: list[ToolCall] | None = None


@dataclass
class ChoiceEvent:
    """Represents a completion event for the AI model."""

    index: int
    message: CompletionMessage
    finish_reason: str = "unknown"
    tool_calls: list[ToolCall] | None = None



import logging
import threading

from opentelemetry import context as context_api
from ..shared import (
    _set_client_attributes,
    _set_request_attributes,
    _set_response_attributes,
    _set_span_attribute,
    is_streaming_response,
    model_as_dict,
    propagate_trace_context,
    set_tools_attributes,
)
from lmnr.sdk.utils import json_dumps
from ..shared.config import Config
from ..utils import (
    _with_chat_telemetry_wrapper,
    dont_throw,
    is_openai_v1,
    should_send_prompts,
)
from lmnr.opentelemetry_lib.tracing.context import (
    get_event_attributes_from_context,
    is_in_litellm_context,
)
from lmnr.opentelemetry_lib.opentelemetry.instrumentation.shared.utils import (
    safe_start_span,
)
from opentelemetry.instrumentation.utils import _SUPPRESS_INSTRUMENTATION_KEY
from opentelemetry.semconv.attributes.error_attributes import ERROR_TYPE
from opentelemetry.trace import Tracer
from opentelemetry.trace.status import Status, StatusCode
from wrapt import ObjectProxy

SPAN_NAME = "openai.chat"

logger = logging.getLogger(__name__)


@_with_chat_telemetry_wrapper
def chat_wrapper(
    tracer: Tracer,
    wrapped,
    instance,
    args,
    kwargs,
):
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
        return wrapped(*args, **kwargs)
    # span needs to be opened and closed manually because the response is a generator

    # LiteLLM calls OpenAI through OpenAI SDK, and to avoid double-instrumentation,
    # we check if we're in a LiteLLM context and return the result directly if so.

    if is_in_litellm_context():
        return wrapped(*args, **kwargs)

    span = safe_start_span(
        name=SPAN_NAME, attributes={"gen_ai.system": "openai"}, span_type="LLM"
    )

    if span is None:
        return wrapped(*args, **kwargs)

    _handle_request(span, kwargs, instance)

    try:
        from lmnr.sdk.rollout_control import is_rollout_mode

        is_rollout = is_rollout_mode()
    except Exception:
        is_rollout = False

    try:
        if is_rollout:
            from lmnr.opentelemetry_lib.opentelemetry.instrumentation.openai.rollout import (
                get_openai_rollout_wrapper,
            )

            rollout_wrapper = get_openai_rollout_wrapper()
            if rollout_wrapper:
                response = rollout_wrapper.wrap_chat_completion(
                    wrapped,
                    instance,
                    args,
                    kwargs,
                    span=span,
                    is_streaming=kwargs.get("stream", False),
                    is_async=False,
                )
            else:
                response = wrapped(*args, **kwargs)
        else:
            response = wrapped(*args, **kwargs)
    except Exception as e:
        span.set_attribute(ERROR_TYPE, e.__class__.__name__)
        attributes = get_event_attributes_from_context()
        span.record_exception(e, attributes=attributes)
        span.set_status(Status(StatusCode.ERROR, str(e)))
        span.end()

        raise

    if is_streaming_response(response):
        if is_openai_v1():
            return ChatStream(
                span,
                response,
                record_raw_response=is_rollout,
            )
        else:
            return _build_from_streaming_response(
                span,
                response,
            )

    _handle_response(
        response,
        span,
        record_raw_response=is_rollout,
    )

    span.end()

    return response


@_with_chat_telemetry_wrapper
async def achat_wrapper(
    tracer: Tracer,
    wrapped,
    instance,
    args,
    kwargs,
):
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
        return await wrapped(*args, **kwargs)

    if is_in_litellm_context():
        return await wrapped(*args, **kwargs)

    span = safe_start_span(
        name=SPAN_NAME, attributes={"gen_ai.system": "openai"}, span_type="LLM"
    )

    if span is None:
        return await wrapped(*args, **kwargs)

    _handle_request(span, kwargs, instance)

    try:
        from lmnr.sdk.rollout_control import is_rollout_mode

        is_rollout = is_rollout_mode()
    except Exception:
        is_rollout = False

    try:
        if is_rollout:
            import inspect

            from lmnr.opentelemetry_lib.opentelemetry.instrumentation.openai.rollout import (
                get_openai_rollout_wrapper,
            )

            rollout_wrapper = get_openai_rollout_wrapper()
            if rollout_wrapper:
                result = rollout_wrapper.wrap_chat_completion(
                    wrapped,
                    instance,
                    args,
                    kwargs,
                    span=span,
                    is_streaming=kwargs.get("stream", False),
                    is_async=True,
                )
                if inspect.iscoroutine(result):
                    response = await result
                elif inspect.isasyncgen(result):
                    response = result
                else:
                    response = result
            else:
                response = await wrapped(*args, **kwargs)
        else:
            response = await wrapped(*args, **kwargs)
    except Exception as e:
        span.set_attribute(ERROR_TYPE, e.__class__.__name__)
        attributes = get_event_attributes_from_context()
        span.record_exception(e, attributes=attributes)
        span.set_status(Status(StatusCode.ERROR, str(e)))
        span.end()

        raise

    if is_streaming_response(response):
        if is_openai_v1():
            return ChatStream(
                span,
                response,
                record_raw_response=is_rollout,
            )
        else:
            return _abuild_from_streaming_response(
                span,
                response,
            )

    _handle_response(
        response,
        span,
        record_raw_response=is_rollout,
    )

    span.end()

    return response


@dont_throw
def _handle_request(span, kwargs, instance):
    _set_request_attributes(span, kwargs, instance)
    _set_client_attributes(span, instance)
    if should_send_prompts():
        _set_prompts(span, kwargs.get("messages"))
        if kwargs.get("functions"):
            set_tools_attributes(span, kwargs.get("functions"))
        elif kwargs.get("tools"):
            set_tools_attributes(span, kwargs.get("tools"))
    if Config.enable_trace_context_propagation:
        propagate_trace_context(span, kwargs)


@dont_throw
def _handle_response(
    response,
    span,
    record_raw_response=False,
):
    if is_openai_v1():
        response_dict = model_as_dict(response)
    else:
        response_dict = response

    _set_response_attributes(span, response_dict)

    if should_send_prompts():
        _set_completions(span, response_dict.get("choices"))

    if record_raw_response:
        try:
            if hasattr(response, "model_dump_json"):
                _set_span_attribute(
                    span, "lmnr.sdk.raw.response", response.model_dump_json()
                )
            else:
                _set_span_attribute(
                    span, "lmnr.sdk.raw.response", json_dumps(response_dict)
                )
        except Exception:
            pass

    return response


def _set_choice_counter_metrics(choice_counter, choices, shared_attributes):
    choice_counter.add(len(choices), attributes=shared_attributes)
    for choice in choices:
        attributes_with_reason = {**shared_attributes}
        if choice.get("finish_reason"):
            attributes_with_reason["llm.response.finish_reason"] = choice.get(
                "finish_reason"
            )
        choice_counter.add(1, attributes=attributes_with_reason)


@dont_throw
def _set_prompts(span, messages):
    if not span.is_recording() or messages is None:
        return

    processed_messages = []
    for msg in messages:
        msg = msg if isinstance(msg, dict) else model_as_dict(msg)
        processed_msg = dict(msg)

        if processed_msg.get("tool_calls"):
            processed_msg["tool_calls"] = [
                model_as_dict(tc) for tc in processed_msg["tool_calls"]
            ]

        processed_messages.append(processed_msg)

    _set_span_attribute(span, "gen_ai.input.messages", json_dumps(processed_messages))


def _set_completions(span, choices):
    if choices is None:
        return

    _set_span_attribute(span, "gen_ai.output.messages", json_dumps(choices))


class ChatStream(ObjectProxy):
    _span = None
    _record_raw_response = False
    _complete_response = None
    _cleanup_completed = False
    _cleanup_lock = None

    def __init__(
        self,
        span,
        response,
        record_raw_response=False,
    ):
        super().__init__(response)

        self._span = span
        self._record_raw_response = record_raw_response
        self._complete_response = {
            "choices": [],
            "model": "",
            "id": "",
            "service_tier": None,
        }

        self._cleanup_completed = False
        self._cleanup_lock = threading.Lock()

    def __del__(self):
        """Cleanup when object is garbage collected"""
        if hasattr(self, "_cleanup_completed") and not self._cleanup_completed:
            self._ensure_cleanup()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        cleanup_exception = None
        try:
            self._ensure_cleanup()
        except Exception as e:
            cleanup_exception = e
            # Don't re-raise to avoid masking original exception

        if hasattr(self.__wrapped__, "__exit__"):
            result = self.__wrapped__.__exit__(exc_type, exc_val, exc_tb)
        else:
            result = None

        if cleanup_exception:
            # Log cleanup exception but don't affect context manager behavior
            logger.debug(
                "Error during ChatStream cleanup in __exit__: %s", cleanup_exception
            )

        return result

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if hasattr(self.__wrapped__, "__aexit__"):
            await self.__wrapped__.__aexit__(exc_type, exc_val, exc_tb)

    def __iter__(self):
        return self

    def __aiter__(self):
        return self

    def __next__(self):
        try:
            chunk = self.__wrapped__.__next__()
        except Exception as e:
            if isinstance(e, StopIteration):
                self._process_complete_response()
            else:
                # Handle cleanup for other exceptions during stream iteration
                self._ensure_cleanup()
                if self._span and self._span.is_recording():
                    self._span.set_status(Status(StatusCode.ERROR, str(e)))
            raise
        else:
            self._process_item(chunk)
            return chunk

    async def __anext__(self):
        try:
            chunk = await self.__wrapped__.__anext__()
        except Exception as e:
            if isinstance(e, StopAsyncIteration):
                self._process_complete_response()
            else:
                # Handle cleanup for other exceptions during stream iteration
                self._ensure_cleanup()
                if self._span and self._span.is_recording():
                    self._span.set_status(Status(StatusCode.ERROR, str(e)))
            raise
        else:
            self._process_item(chunk)
            return chunk

    def _process_item(self, item):
        self._span.add_event(name="llm.content.completion.chunk")
        self._complete_response["id"] = item.id if hasattr(item, "id") else ""
        self._complete_response["service_tier"] = (
            item.service_tier if hasattr(item, "service_tier") else ""
        )

        _accumulate_stream_items(item, self._complete_response)

    @dont_throw
    def _process_complete_response(self):
        _set_response_attributes(self._span, self._complete_response)
        if should_send_prompts():
            _set_completions(self._span, self._complete_response.get("choices"))

        if self._record_raw_response:
            try:
                _set_span_attribute(
                    self._span,
                    "lmnr.sdk.raw.response",
                    json_dumps(self._complete_response),
                )
            except Exception:
                pass

        self._span.set_status(Status(StatusCode.OK))
        self._span.end()
        self._cleanup_completed = True

    @dont_throw
    def _ensure_cleanup(self):
        """Thread-safe cleanup method that handles different cleanup scenarios"""
        with self._cleanup_lock:
            if self._cleanup_completed:
                logger.debug("ChatStream cleanup already completed, skipping")
                return

            try:
                logger.debug("Starting ChatStream cleanup")

                # Set span status and close it
                if self._span and self._span.is_recording():
                    self._span.set_status(Status(StatusCode.OK))
                    self._span.end()
                    logger.debug("ChatStream span closed successfully")

                self._cleanup_completed = True
                logger.debug("ChatStream cleanup completed successfully")

            except Exception as e:
                # Log cleanup errors but don't propagate to avoid masking original issues
                logger.debug("Error during ChatStream cleanup: %s", str(e))

                # Still try to close the span even if metrics recording failed
                try:
                    if self._span and self._span.is_recording():
                        self._span.set_status(
                            Status(StatusCode.ERROR, "Cleanup failed")
                        )
                        self._span.end()
                    self._cleanup_completed = True
                except Exception:
                    # Final fallback - just mark as completed to prevent infinite loops
                    self._cleanup_completed = True


# Backward compatibility with OpenAI v0


@dont_throw
def _build_from_streaming_response(
    span,
    response,
):
    complete_response = {"choices": [], "model": "", "id": "", "service_tier": None}

    for item in response:
        span.add_event(name="llm.content.completion.chunk")

        item_to_yield = item

        _accumulate_stream_items(item, complete_response)

        yield item_to_yield

    _set_response_attributes(span, complete_response)
    if should_send_prompts():
        _set_completions(span, complete_response.get("choices"))

    span.set_status(Status(StatusCode.OK))
    span.end()


@dont_throw
async def _abuild_from_streaming_response(
    span,
    response,
):
    complete_response = {"choices": [], "model": "", "id": "", "service_tier": None}

    async for item in response:
        span.add_event(name="llm.content.completion.chunk")

        item_to_yield = item

        _accumulate_stream_items(item, complete_response)

        yield item_to_yield

    _set_response_attributes(span, complete_response)
    if should_send_prompts():
        _set_completions(span, complete_response.get("choices"))

    span.set_status(Status(StatusCode.OK))
    span.end()


def _accumulate_stream_items(item, complete_response):
    if is_openai_v1():
        item = model_as_dict(item)

    complete_response["model"] = item.get("model")
    complete_response["id"] = item.get("id")
    complete_response["service_tier"] = item.get("service_tier")
    if item.get("created"):
        complete_response["created"] = item.get("created")
    if "object" not in complete_response:
        complete_response["object"] = "chat.completion"

    # capture usage information from the last stream chunks
    if item.get("usage"):
        complete_response["usage"] = item.get("usage")
    elif item.get("choices") and item["choices"][0].get("usage"):
        # Some LLM providers like moonshot mistakenly place token usage information within choices[0], handle this.
        complete_response["usage"] = item["choices"][0].get("usage")

    # prompt filter results
    if item.get("prompt_filter_results"):
        complete_response["prompt_filter_results"] = item.get("prompt_filter_results")

    for choice in item.get("choices") or []:
        index = choice.get("index")
        if len(complete_response.get("choices")) <= index:
            complete_response["choices"].append(
                {"index": index, "message": {"content": "", "role": ""}}
            )
        complete_choice = complete_response.get("choices")[index]
        if choice.get("finish_reason"):
            complete_choice["finish_reason"] = choice.get("finish_reason")
        if choice.get("content_filter_results"):
            complete_choice["content_filter_results"] = choice.get(
                "content_filter_results"
            )

        delta = choice.get("delta")

        if delta and delta.get("content"):
            complete_choice["message"]["content"] += delta.get("content")

        if delta and delta.get("role"):
            complete_choice["message"]["role"] = delta.get("role")
        if delta and delta.get("tool_calls"):
            tool_calls = delta.get("tool_calls")
            if not isinstance(tool_calls, list) or len(tool_calls) == 0:
                continue

            if not complete_choice["message"].get("tool_calls"):
                complete_choice["message"]["tool_calls"] = []

            for tool_call in tool_calls:
                i = int(tool_call["index"])
                if len(complete_choice["message"]["tool_calls"]) <= i:
                    complete_choice["message"]["tool_calls"].append(
                        {
                            "id": "",
                            "type": "function",
                            "function": {"name": "", "arguments": ""},
                        }
                    )

                span_tool_call = complete_choice["message"]["tool_calls"][i]
                span_function = span_tool_call["function"]
                tool_call_function = tool_call.get("function")

                if tool_call.get("id"):
                    span_tool_call["id"] = tool_call.get("id")
                if tool_call_function and tool_call_function.get("name"):
                    span_function["name"] = tool_call_function.get("name")
                if tool_call_function and tool_call_function.get("arguments"):
                    span_function["arguments"] += tool_call_function.get("arguments")



"""OpenTelemetry Google Generative AI API instrumentation"""

import json
import logging
import os
from collections import defaultdict
from typing import AsyncGenerator, Callable, Collection, Generator

from google.genai import types
from opentelemetry import context as context_api
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry.instrumentation.utils import _SUPPRESS_INSTRUMENTATION_KEY, unwrap
from opentelemetry.semconv._incubating.attributes import gen_ai_attributes
from opentelemetry.semconv.attributes.error_attributes import ERROR_TYPE
from opentelemetry.trace import Span, Status, StatusCode, Tracer, get_tracer
from wrapt import wrap_function_wrapper

from lmnr.opentelemetry_lib.opentelemetry.instrumentation.shared.utils import (
    safe_start_span,
)
from lmnr.opentelemetry_lib.tracing.context import (
    get_event_attributes_from_context,
)
from lmnr.sdk.laminar import Laminar
from lmnr.sdk.utils import json_dumps

from .config import (
    Config,
)
from .schema_utils import SchemaJSONEncoder, process_schema
from .utils import (
    content_union_to_dict,
    dont_throw,
    get_content,
    merge_text_parts,
    process_content_union,
    process_stream_chunk,
    set_span_attribute,
    to_dict,
    with_tracer_wrapper,
)

logger = logging.getLogger(__name__)

_instruments = ("google-genai >= 1.0.0",)

WRAPPED_METHODS = [
    {
        "package": "google.genai.models",
        "object": "Models",
        "method": "generate_content",
        "span_name": "gemini.generate_content",
        "is_streaming": False,
        "is_async": False,
    },
    {
        "package": "google.genai.models",
        "object": "AsyncModels",
        "method": "generate_content",
        "span_name": "gemini.generate_content",
        "is_streaming": False,
        "is_async": True,
    },
    {
        "package": "google.genai.models",
        "object": "Models",
        "method": "generate_content_stream",
        "span_name": "gemini.generate_content_stream",
        "is_streaming": True,
        "is_async": False,
    },
    {
        "package": "google.genai.models",
        "object": "AsyncModels",
        "method": "generate_content_stream",
        "span_name": "gemini.generate_content_stream",
        "is_streaming": True,
        "is_async": True,
    },
]


def should_send_prompts():
    return (
        os.getenv("LAMINAR_TRACE_CONTENT") or "true"
    ).lower() == "true" or context_api.get_value("override_enable_content_tracing")


@dont_throw
def _set_request_attributes(span, args, kwargs):
    config_dict = to_dict(kwargs.get("config", {}))
    set_span_attribute(
        span, gen_ai_attributes.GEN_AI_REQUEST_MODEL, kwargs.get("model")
    )
    set_span_attribute(
        span,
        gen_ai_attributes.GEN_AI_REQUEST_TEMPERATURE,
        config_dict.get("temperature"),
    )
    set_span_attribute(
        span, gen_ai_attributes.GEN_AI_REQUEST_TOP_P, config_dict.get("top_p")
    )
    set_span_attribute(
        span, gen_ai_attributes.GEN_AI_REQUEST_TOP_K, config_dict.get("top_k")
    )
    set_span_attribute(
        span,
        gen_ai_attributes.GEN_AI_REQUEST_CHOICE_COUNT,
        config_dict.get("candidate_count"),
    )
    set_span_attribute(
        span,
        gen_ai_attributes.GEN_AI_REQUEST_MAX_TOKENS,
        config_dict.get("max_output_tokens"),
    )
    set_span_attribute(
        span,
        gen_ai_attributes.GEN_AI_REQUEST_STOP_SEQUENCES,
        config_dict.get("stop_sequences"),
    )
    set_span_attribute(
        span,
        gen_ai_attributes.GEN_AI_REQUEST_FREQUENCY_PENALTY,
        config_dict.get("frequency_penalty"),
    )
    set_span_attribute(
        span,
        gen_ai_attributes.GEN_AI_REQUEST_PRESENCE_PENALTY,
        config_dict.get("presence_penalty"),
    )
    set_span_attribute(
        span, gen_ai_attributes.GEN_AI_REQUEST_SEED, config_dict.get("seed")
    )

    if schema := config_dict.get("response_schema"):
        try:
            set_span_attribute(
                span,
                "gen_ai.request.structured_output_schema",
                json.dumps(process_schema(schema), cls=SchemaJSONEncoder),
            )
        except Exception:
            pass
    elif json_schema := config_dict.get("response_json_schema"):
        try:
            set_span_attribute(
                span,
                "gen_ai.request.structured_output_schema",
                json_dumps(json_schema),
            )
        except Exception:
            pass

    tools: list[types.FunctionDeclaration] = []
    arg_tools = config_dict.get("tools", kwargs.get("tools"))
    if arg_tools:
        for tool in arg_tools:
            if isinstance(tool, types.Tool):
                tools.extend(tool.function_declarations or [])
            elif isinstance(tool, dict) and isinstance(
                tool.get("function_declarations"), list
            ):
                tools.extend(tool.get("function_declarations", []))
            elif isinstance(tool, Callable):
                tools.append(types.FunctionDeclaration.from_callable(tool))

    if should_send_prompts():
        messages = []
        system_instruction = config_dict.get("system_instruction")
        if system_instruction:
            msg = content_union_to_dict(system_instruction, default_role="system")
            msg["role"] = "system"
            messages.append(msg)

        contents = kwargs.get("contents", [])
        if not isinstance(contents, list):
            contents = [contents]
        for content in contents:
            messages.append(content_union_to_dict(content))

        set_span_attribute(span, "gen_ai.input.messages", json_dumps(messages))
    if tools:
        span.set_attribute(
            "gen_ai.tool.definitions",
            json_dumps([to_dict(tool) for tool in tools]),
        )


@dont_throw
def _set_response_attributes(span, response: types.GenerateContentResponse):
    candidates = response.candidates or []
    set_span_attribute(
        span, gen_ai_attributes.GEN_AI_RESPONSE_ID, to_dict(response).get("response_id")
    )
    set_span_attribute(
        span,
        gen_ai_attributes.GEN_AI_RESPONSE_MODEL,
        to_dict(response).get("model_version"),
    )

    if response.usage_metadata:
        usage_dict = to_dict(response.usage_metadata)
        candidates_token_count = usage_dict.get("candidates_token_count")
        # unlike OpenAI, and unlike input cached tokens, thinking tokens are
        # not counted as part of candidates token count, so we need to add them
        # separately for consistency with other instrumentations
        thoughts_token_count = usage_dict.get("thoughts_token_count")
        output_token_count = (
            (candidates_token_count or 0) + (thoughts_token_count or 0)
            if candidates_token_count is not None or thoughts_token_count is not None
            else None
        )
        set_span_attribute(
            span,
            gen_ai_attributes.GEN_AI_USAGE_INPUT_TOKENS,
            usage_dict.get("prompt_token_count"),
        )
        set_span_attribute(
            span,
            gen_ai_attributes.GEN_AI_USAGE_OUTPUT_TOKENS,
            output_token_count,
        )
        set_span_attribute(
            span,
            "llm.usage.total_tokens",
            usage_dict.get("total_token_count"),
        )
        set_span_attribute(
            span,
            "gen_ai.usage.cache_read_input_tokens",
            usage_dict.get("cached_content_token_count"),
        )
        set_span_attribute(
            span,
            "gen_ai.usage.reasoning_tokens",
            thoughts_token_count,
        )

    if should_send_prompts():
        set_span_attribute(span, "gen_ai.completion.0.role", "model")
        candidates_list = candidates if isinstance(candidates, list) else [candidates]
        i = 0
        for candidate in candidates_list:
            has_content = False
            processed_content = process_content_union(candidate.content)
            content_payload = get_content(processed_content)
            if isinstance(content_payload, dict):
                content_payload = [content_payload]

            set_span_attribute(span, f"gen_ai.completion.{i}.role", "model")
            if content_payload:
                has_content = True
                set_span_attribute(
                    span,
                    f"gen_ai.completion.{i}.content",
                    (
                        content_payload
                        if isinstance(content_payload, str)
                        else json_dumps(content_payload)
                    ),
                )
            blocks = (
                processed_content
                if isinstance(processed_content, list)
                else [processed_content]
            )

            tool_call_index = 0
            for block in blocks:
                block_dict = to_dict(block)
                if not block_dict.get("function_call"):
                    continue
                function_call = to_dict(block_dict.get("function_call", {}))
                has_content = True
                set_span_attribute(
                    span,
                    f"gen_ai.completion.{i}.tool_calls.{tool_call_index}.name",
                    function_call.get("name"),
                )
                set_span_attribute(
                    span,
                    f"gen_ai.completion.{i}.tool_calls.{tool_call_index}.id",
                    (
                        function_call.get("id")
                        if function_call.get("id") is not None
                        else function_call.get("name")
                    ),  # google genai doesn't support tool call ids
                )
                set_span_attribute(
                    span,
                    f"gen_ai.completion.{i}.tool_calls.{tool_call_index}.arguments",
                    json_dumps(function_call.get("arguments")),
                )
                tool_call_index += 1
            if has_content:
                i += 1


@dont_throw
def _set_raw_response_attribute(
    span, response: types.GenerateContentResponse, record_raw_response: bool = False
):
    set_span_attribute(
        span,
        "gen_ai.output.messages",
        json_dumps(
            [
                candidate.model_dump(mode="json", exclude_unset=True)
                for candidate in (response.candidates or [])
            ]
        ),
    )
    if record_raw_response:
        try:
            set_span_attribute(
                span,
                "lmnr.sdk.raw.response",
                response.model_dump_json(),
            )
        except Exception:
            logger.debug("Failed to set lmnr.sdk.raw.response attribute", exc_info=True)


@dont_throw
def _build_from_streaming_response(
    span: Span,
    response: Generator[types.GenerateContentResponse, None, None],
    record_raw_response: bool = False,
) -> Generator[types.GenerateContentResponse, None, None]:
    final_parts = []
    role = "model"
    aggregated_usage_metadata = defaultdict(int)
    model_version = None
    for chunk in response:
        try:
            span.add_event("llm.content.completion.chunk")
        except Exception:
            pass
        # Important: do all processing in a separate sync function, that is
        # wrapped in @dont_throw. If we did it here, the @dont_throw on top of
        # this function would not be able to catch the errors, as they are
        # raised later, after the generator is returned, and when it is being
        # consumed.
        chunk_result = process_stream_chunk(
            chunk,
            role,
            model_version,
            aggregated_usage_metadata,
            final_parts,
        )
        # even though process_stream_chunk can't return None, the result can be
        # None, if the processing throws an error (see @dont_throw)
        if chunk_result:
            role = chunk_result["role"]
            model_version = chunk_result["model_version"]
        yield chunk

    try:
        compound_response = types.GenerateContentResponse(
            candidates=[
                {
                    "content": {
                        "parts": merge_text_parts(final_parts),
                        "role": role,
                    },
                }
            ],
            usage_metadata=types.GenerateContentResponseUsageMetadataDict(
                **aggregated_usage_metadata
            ),
            model_version=model_version,
        )
        if span.is_recording():
            _set_raw_response_attribute(
                span, compound_response, record_raw_response=record_raw_response
            )
            _set_response_attributes(span, compound_response)
    finally:
        if span.is_recording():
            span.end()


@dont_throw
async def _abuild_from_streaming_response(
    span: Span,
    response: AsyncGenerator[types.GenerateContentResponse, None],
    record_raw_response: bool = False,
) -> AsyncGenerator[types.GenerateContentResponse, None]:
    final_parts = []
    role = "model"
    aggregated_usage_metadata = defaultdict(int)
    model_version = None
    async for chunk in response:
        try:
            span.add_event("llm.content.completion.chunk")
        except Exception:
            pass
        # Important: do all processing in a separate sync function, that is
        # wrapped in @dont_throw. If we did it here, the @dont_throw on top of
        # this function would not be able to catch the errors, as they are
        # raised later, after the generator is returned, and when it is being
        # consumed.
        chunk_result = process_stream_chunk(
            chunk,
            role,
            model_version,
            aggregated_usage_metadata,
            final_parts,
        )
        # even though process_stream_chunk can't return None, the result can be
        # None, if the processing throws an error (see @dont_throw)
        if chunk_result:
            role = chunk_result["role"]
            model_version = chunk_result["model_version"]
        yield chunk

    try:
        compound_response = types.GenerateContentResponse(
            candidates=[
                {
                    "content": {
                        "parts": merge_text_parts(final_parts),
                        "role": role,
                    },
                }
            ],
            usage_metadata=types.GenerateContentResponseUsageMetadataDict(
                **aggregated_usage_metadata
            ),
            model_version=model_version,
        )
        if span.is_recording():
            _set_raw_response_attribute(
                span, compound_response, record_raw_response=record_raw_response
            )
            _set_response_attributes(span, compound_response)
    finally:
        if span.is_recording():
            span.end()


@with_tracer_wrapper
def _wrap(tracer: Tracer, to_wrap, wrapped, instance, args, kwargs):
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
        return wrapped(*args, **kwargs)

    span = safe_start_span(
        name=to_wrap.get("span_name"),
        attributes={"gen_ai.system": "gemini"},
        span_type="LLM",
    )
    if not span:
        logger.warning("Failed to start span for google genai")
        return wrapped(*args, **kwargs)

    _set_request_attributes(span, args, kwargs)

    try:
        # Check for rollout mode and apply caching/overrides
        from lmnr.sdk.rollout_control import is_rollout_mode

        is_rollout = is_rollout_mode()
    except Exception:
        is_rollout = False

    try:
        if is_rollout:
            from lmnr.opentelemetry_lib.opentelemetry.instrumentation.google_genai.rollout import (
                get_google_genai_rollout_wrapper,
            )

            rollout_wrapper = get_google_genai_rollout_wrapper()
            if rollout_wrapper:
                with Laminar.use_span(span):
                    response = rollout_wrapper.wrap_generate_content(
                        wrapped,
                        instance,
                        args,
                        kwargs,
                        is_streaming=to_wrap.get("is_streaming", False),
                        is_async=False,
                    )
            else:
                response = wrapped(*args, **kwargs)
        else:
            response = wrapped(*args, **kwargs)

        if to_wrap.get("is_streaming"):
            return _build_from_streaming_response(
                span, response, record_raw_response=is_rollout
            )
        if span.is_recording():
            _set_raw_response_attribute(span, response, record_raw_response=is_rollout)
            _set_response_attributes(span, response)
        span.end()
        return response
    except Exception as e:
        attributes = get_event_attributes_from_context()
        span.set_attribute(ERROR_TYPE, e.__class__.__name__)
        span.record_exception(e, attributes=attributes)
        span.set_status(Status(StatusCode.ERROR, str(e)))
        span.end()
        raise


@with_tracer_wrapper
async def _awrap(tracer: Tracer, to_wrap, wrapped, instance, args, kwargs):
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
        return await wrapped(*args, **kwargs)

    span = safe_start_span(
        name=to_wrap.get("span_name"),
        attributes={"gen_ai.system": "gemini"},
        span_type="LLM",
    )
    if not span:
        logger.warning("Failed to start span for async google genai")
        return await wrapped(*args, **kwargs)

    _set_request_attributes(span, args, kwargs)

    try:
        # Check for rollout mode and apply caching/overrides
        from lmnr.sdk.rollout_control import is_rollout_mode

        is_rollout = is_rollout_mode()
    except Exception:
        is_rollout = False

    try:
        if is_rollout:
            from lmnr.opentelemetry_lib.opentelemetry.instrumentation.google_genai.rollout import (
                get_google_genai_rollout_wrapper,
            )

            rollout_wrapper = get_google_genai_rollout_wrapper()
            if rollout_wrapper:
                # For async, we need to handle both sync and async wrapped functions
                # In rollout mode, wrapped might return cached (sync) or live (async)
                with Laminar.use_span(span):
                    result = rollout_wrapper.wrap_generate_content(
                        wrapped,
                        instance,
                        args,
                        kwargs,
                        is_streaming=to_wrap.get("is_streaming", False),
                        is_async=True,
                    )
                # If result is a coroutine or async generator, await/iterate it
                import inspect

                if inspect.iscoroutine(result):
                    response = await result
                elif inspect.isasyncgen(result):
                    # It's an async generator (cached streaming response)
                    response = result
                else:
                    # It's a sync response (cached non-streaming)
                    response = result
            else:
                response = await wrapped(*args, **kwargs)
        else:
            response = await wrapped(*args, **kwargs)

        if to_wrap.get("is_streaming"):
            return _abuild_from_streaming_response(
                span, response, record_raw_response=is_rollout
            )
        else:
            if span.is_recording():
                _set_raw_response_attribute(
                    span, response, record_raw_response=is_rollout
                )
                _set_response_attributes(span, response)

            span.end()
            return response
    except Exception as e:
        attributes = get_event_attributes_from_context()
        span.set_attribute(ERROR_TYPE, e.__class__.__name__)
        span.record_exception(e, attributes=attributes)
        span.set_status(Status(StatusCode.ERROR, str(e)))
        span.end()
        raise


class GoogleGenAiSdkInstrumentor(BaseInstrumentor):
    """An instrumentor for Google GenAI's client library."""

    def __init__(
        self,
        exception_logger=None,
        upload_base64_image=None,
        convert_image_to_openai_format=True,
    ):
        super().__init__()
        Config.exception_logger = exception_logger
        Config.upload_base64_image = upload_base64_image
        Config.convert_image_to_openai_format = convert_image_to_openai_format

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        tracer_provider = kwargs.get("tracer_provider")
        tracer = get_tracer(__name__, "0.0.1a1", tracer_provider)

        for wrapped_method in WRAPPED_METHODS:
            wrap_function_wrapper(
                wrapped_method.get("package"),
                f"{wrapped_method.get('object')}.{wrapped_method.get('method')}",
                (
                    _awrap(tracer, wrapped_method)
                    if wrapped_method.get("is_async")
                    else _wrap(tracer, wrapped_method)
                ),
            )

    def _uninstrument(self, **kwargs):
        for wrapped_method in WRAPPED_METHODS:
            unwrap(
                f"{wrapped_method.get('package')}.{wrapped_method.get('object')}",
                wrapped_method.get("method"),
            )



from lmnr.opentelemetry_lib.tracing.context import get_event_attributes_from_context
from ..shared import _set_span_attribute
from opentelemetry.semconv.attributes.error_attributes import ERROR_TYPE
from opentelemetry.semconv._incubating.attributes.gen_ai_attributes import (
    GEN_AI_USAGE_INPUT_TOKENS,
    GEN_AI_USAGE_OUTPUT_TOKENS,
)
from opentelemetry.trace import Status, StatusCode
from typing_extensions import override

from openai import AssistantEventHandler


class EventHandlerWrapper(AssistantEventHandler):
    _current_text_index = 0
    _prompt_tokens = 0
    _completion_tokens = 0

    def __init__(self, original_handler, span):
        super().__init__()
        self._original_handler = original_handler
        self._span = span

    @override
    def on_end(self):
        _set_span_attribute(
            self._span,
            GEN_AI_USAGE_INPUT_TOKENS,
            self._prompt_tokens,
        )
        _set_span_attribute(
            self._span,
            GEN_AI_USAGE_OUTPUT_TOKENS,
            self._completion_tokens,
        )
        self._original_handler.on_end()
        self._span.end()

    @override
    def on_event(self, event):
        self._original_handler.on_event(event)

    @override
    def on_run_step_created(self, run_step):
        self._original_handler.on_run_step_created(run_step)

    @override
    def on_run_step_delta(self, delta, snapshot):
        self._original_handler.on_run_step_delta(delta, snapshot)

    @override
    def on_run_step_done(self, run_step):
        if run_step.usage:
            self._prompt_tokens += run_step.usage.prompt_tokens
            self._completion_tokens += run_step.usage.completion_tokens
        self._original_handler.on_run_step_done(run_step)

    @override
    def on_tool_call_created(self, tool_call):
        self._original_handler.on_tool_call_created(tool_call)

    @override
    def on_tool_call_delta(self, delta, snapshot):
        self._original_handler.on_tool_call_delta(delta, snapshot)

    @override
    def on_tool_call_done(self, tool_call):
        self._original_handler.on_tool_call_done(tool_call)

    @override
    def on_exception(self, exception: Exception):
        self._span.set_attribute(ERROR_TYPE, exception.__class__.__name__)
        self._span.record_exception(
            exception, attributes=get_event_attributes_from_context()
        )
        self._span.set_status(Status(StatusCode.ERROR, str(exception)))
        self._original_handler.on_exception(exception)

    @override
    def on_timeout(self):
        self._original_handler.on_timeout()

    @override
    def on_message_created(self, message):
        self._original_handler.on_message_created(message)

    @override
    def on_message_delta(self, delta, snapshot):
        self._original_handler.on_message_delta(delta, snapshot)

    @override
    def on_message_done(self, message):
        _set_span_attribute(
            self._span,
            f"gen_ai.response.{self._current_text_index}.id",
            message.id,
        )
        self._original_handler.on_message_done(message)
        self._current_text_index += 1

    @override
    def on_text_created(self, text):
        self._original_handler.on_text_created(text)

    @override
    def on_text_delta(self, delta, snapshot):
        self._original_handler.on_text_delta(delta, snapshot)

    @override
    def on_text_done(self, text):
        self._original_handler.on_text_done(text)
        _set_span_attribute(
            self._span,
            f"gen_ai.completion.{self._current_text_index}.role",
            "assistant",
        )
        _set_span_attribute(
            self._span,
            f"gen_ai.completion.{self._current_text_index}.content",
            text.value,
        )

    @override
    def on_image_file_done(self, image_file):
        self._original_handler.on_image_file_done(image_file)



import logging

from opentelemetry import context as context_api

from lmnr.opentelemetry_lib.tracing.context import get_event_attributes_from_context
from ..shared import (
    _set_client_attributes,
    _set_request_attributes,
    _set_response_attributes,
    _set_span_attribute,
    model_as_dict,
    propagate_trace_context,
)
from ..shared.config import Config
from ..utils import (
    _with_embeddings_telemetry_wrapper,
    dont_throw,
    is_openai_v1,
    should_send_prompts,
)
from lmnr.opentelemetry_lib.opentelemetry.instrumentation.shared.utils import (
    safe_start_span,
)
from lmnr.sdk.utils import json_dumps
from opentelemetry.instrumentation.utils import _SUPPRESS_INSTRUMENTATION_KEY
from opentelemetry.semconv.attributes.error_attributes import ERROR_TYPE
from opentelemetry.trace import Status, StatusCode

SPAN_NAME = "openai.embeddings"
logger = logging.getLogger(__name__)


@_with_embeddings_telemetry_wrapper
def embeddings_wrapper(
    tracer,
    wrapped,
    instance,
    args,
    kwargs,
):
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
        return wrapped(*args, **kwargs)

    span = safe_start_span(
        name=SPAN_NAME,
        attributes={"gen_ai.system": "openai"},
        span_type="LLM",
    )
    if span is None:
        return wrapped(*args, **kwargs)

    _handle_request(span, kwargs, instance)

    try:
        response = wrapped(*args, **kwargs)
        _handle_response(response, span)
        return response
    except Exception as e:  # pylint: disable=broad-except
        attributes = {"error.type": e.__class__.__name__}

        span.set_attribute("error.type", e.__class__.__name__)
        attributes = get_event_attributes_from_context()
        span.record_exception(e, attributes=attributes)
        span.set_status(Status(StatusCode.ERROR, str(e)))
        raise
    finally:
        span.end()


@_with_embeddings_telemetry_wrapper
async def aembeddings_wrapper(
    tracer,
    wrapped,
    instance,
    args,
    kwargs,
):
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
        return await wrapped(*args, **kwargs)

    span = safe_start_span(
        name=SPAN_NAME,
        attributes={"gen_ai.system": "openai"},
        span_type="LLM",
    )
    if span is None:
        return await wrapped(*args, **kwargs)

    _handle_request(span, kwargs, instance)

    try:
        response = await wrapped(*args, **kwargs)
        _handle_response(response, span)
        return response
    except Exception as e:  # pylint: disable=broad-except
        attributes = {
            "error.type": e.__class__.__name__,
        }

        span.set_attribute(ERROR_TYPE, e.__class__.__name__)
        attributes = get_event_attributes_from_context()
        span.record_exception(e, attributes=attributes)
        span.set_status(Status(StatusCode.ERROR, str(e)))
        raise
    finally:
        span.end()


@dont_throw
def _handle_request(span, kwargs, instance):
    _set_request_attributes(span, kwargs, instance)

    if should_send_prompts():
        _set_prompts(span, kwargs.get("input"))

    _set_client_attributes(span, instance)

    if Config.enable_trace_context_propagation:
        propagate_trace_context(span, kwargs)


@dont_throw
def _handle_response(
    response,
    span,
):
    if is_openai_v1():
        response_dict = model_as_dict(response)
    else:
        response_dict = response
    # span attributes
    _set_response_attributes(span, response_dict)


def _set_prompts(span, prompt):
    if not span.is_recording() or not prompt:
        return

    if isinstance(prompt, list):
        messages = [{"content": p} for p in prompt]
    else:
        messages = [{"content": prompt}]
    _set_span_attribute(span, "gen_ai.input.messages", json_dumps(messages))
